# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for dot_general."""
import copy
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import config
from aqt.jax.v2 import int_numerics
from aqt.jax.v2 import stochastic_rounding
import aqt.jax.v2.aqt_dot_general as aqt
import aqt.jax.v2.examples.mnist as aqt_mnist
from aqt.jax.v2.flax import aqt_flax
import flax.linen.linear as fl
import jax
import jax.numpy as jnp
import numpy as np
import scipy.stats


def test_jaxpr_dtype(f, cfgs: list[config.DotGeneralRaw], float_dtype):
  """Tests whether dot_generals in f conform to dtypes inside of cfgs."""

  def jaxpr_to_trityp(jaxpr):
    for eq in jaxpr.eqns:
      # This is where dot_generals hide when one uses custom vjp
      if "fun_jaxpr" in eq.params.keys():
        for rec in jaxpr_to_trityp(eq.params["fun_jaxpr"]):
          yield rec

      if eq.primitive.name == "dot_general":
        [lhs, rhs] = eq.invars
        [out] = eq.outvars
        trityp = (lhs.aval, rhs.aval, out.aval)
        yield trityp

  f_jaxpr = jax.make_jaxpr(f)()
  trityps = [trityp for trityp in jaxpr_to_trityp(f_jaxpr)]
  assert len(trityps) == len(cfgs)
  for (lhs_sa, rhs_sa, out_sa), cfg in zip(trityps, cfgs):
    # If cfg has None, the type is inherited from the arguments' type.
    in_dtype = cfg.dg_in_dtype or float_dtype
    out_dtype = cfg.dg_accumulator_dtype or float_dtype
    assert lhs_sa.dtype == in_dtype
    assert rhs_sa.dtype == in_dtype
    assert out_sa.dtype == out_dtype


def rand_unif(shape, maxval, seed, dtype=jnp.float32):
  key = jax.random.PRNGKey(seed)
  return jax.random.uniform(
      key=key, shape=shape, minval=-maxval, maxval=maxval, dtype=dtype
  )


# The main test strategy is :
# - Test that FQ is sensible (improve fq_noise_test)
#   - Quantization noise (rounding error) is of proper value
#   - Cliping is correct
#   - Stochastic noise is added correctly.
# - Test that we really use int8/int4/bool in XLA/HLO
# - Test of fq vs dot_general equivalence in po2
# - Test that it all applies to gradients
#
#
# TODO(lew): Tests are a bit incomplete. What we need:
# On top of that we shell test:
#  - Gradinent is 1 in the range, 0 outside the abs-max range.


def test_eq(name, a, b):
  # TODO(lew): use library function.
  mean_err = jnp.mean(jnp.abs(a - b))
  if mean_err != 0.0:
    print("mean_err =", mean_err)
    print(a.shape)
    print(a[:3, :3])
    print(b.shape)
    print(b[:3, :3])
    print(f"FAIL: {name}")
    assert False


def fqt_param_dict(s, use_fwd_quant, **kwargs):
  return dict(
      cfg=config.fully_quantized(
          use_fwd_quant=use_fwd_quant,
          use_stochastic_rounding=False,
          **kwargs,
      ),
      seed=s,
  )


class AqtDotGeneralResearchTest(parameterized.TestCase):

  def test_empty(self):
    t = np.random.normal(size=6).reshape((2, 3))
    np.testing.assert_array_equal(t, t)

  @parameterized.product(
      preserve_zero=[True, False],
      prec=[1, 2, 4, 8],
      v=[0.1, 1000.0],
      seed=list(range(10)),
  )
  def test_fq_noise(self, preserve_zero, prec, v, seed):
    key = jax.random.PRNGKey(seed)
    cfg = config.tensor_make(prec)
    if isinstance(cfg.numerics, int_numerics.IntNumerics):
      cfg.numerics = cfg.numerics.replace(preserve_zero=preserve_zero)
    cfg.calib_shared_axes = (0,)
    sample_size = 10000
    shape = (sample_size,)
    a = jax.random.uniform(key, shape, minval=-v, maxval=v)
    context = aqt.Context(key=None, train_step=None)
    a_fq = aqt.make_fake_quant(cfg)(a, context)
    bucket_noise = a_fq - a  #  ~ U(-bucket_size/2, bucket_size/2)
    bucket_count = (2**prec - 1) if preserve_zero else (2**prec)
    bucket_size = (v * 2) / bucket_count
    noise = bucket_noise / bucket_size + 0.5  # ~U(0, 1)
    pvalue = scipy.stats.kstest(noise, "uniform").pvalue
    assert pvalue > 0.01

  def test_stochastic_rounding_noise(self):
    repeats = 1000
    num_values = 100000
    shape = (repeats, num_values)

    def assert_clt(noise: jnp.ndarray):
      # Test if the sample mean of the noise conforms to Central Limit Theorem

      # std of a uniform distribution [-0.5, 0.5]
      noise_std = 1 / jnp.sqrt(12.0)
      noise_mean = jnp.mean(noise, axis=1)
      assert jnp.all(noise_mean * jnp.sqrt(num_values) < 4 * noise_std)

    # jax.uniform implementation
    jax_uniform_noise = jax.random.uniform(jax.random.PRNGKey(10), shape) - 0.5
    assert_clt(jax_uniform_noise)
    # customized more efficient implementation
    custom_1_noise = stochastic_rounding.random_centered_uniform(
        shape, jax.random.PRNGKey(11)
    )
    assert_clt(custom_1_noise)

  @parameterized.parameters([
      dict(bits=1),
  ])
  def test_fake_quant(
      self,
      bits=4,
      maxval=10.0,
      shape=(20, 1),
  ):
    cfg = config.tensor_make(bits)
    cfg.po2_scale = True
    cfg.calib_shared_axes = (0,)
    x = jnp.linspace(-maxval, maxval, num=shape[0]).reshape(shape)
    grad = jnp.ones(shape) * 12345.0
    context = aqt.Context(key=None, train_step=None)
    x_fq, backprop = jax.vjp(aqt.make_fake_quant(cfg), x, context)
    gx_fq = backprop(grad)
    # print(f"x     =\n{x}")
    # print(f"x_fq  =\n{x_fq}")
    # print(f"grad  =\n{grad}")
    # print(f"gx_fq =\n{gx_fq}")
    del x
    del x_fq
    del grad
    del gx_fq

    # TODO(lew): test

  @parameterized.parameters([
      dict(cfg=config.config_v3(fwd_bits=3, dlhs_bits=4, drhs_bits=5)),
      dict(cfg=config.dot_general_make(None, None)),
      dict(cfg=config.dot_general_make(1, 1)),
      dict(cfg=config.dot_general_make(1, 2)),
      dict(cfg=config.dot_general_make(2, 1)),
      dict(cfg=config.dot_general_make(2, 2)),
      dict(cfg=config.dot_general_make(8, 8)),
      dict(cfg=config.dot_general_make(8, 8), clip_gradient=True),
      dict(
          cfg=config.dot_general_make(8, 8, dlhs_local_aqt=config.LocalAqt(2))
      ),
      dict(
          cfg=config.dot_general_make(8, 8, drhs_local_aqt=config.LocalAqt(2))
      ),
      # That test could fail numerically because bf16
      # can't keep in the product of int8*int8 accurately.
      # It just so happens that this test does not fail but others do.
      # We do this test anyway, to catch jax-compilation-time errors.
      dict(cfg=config.dot_general_make(2, 2), dtype=jnp.bfloat16),
      dict(cfg=config.dot_general_make(8, 8), dtype=jnp.bfloat16),
      dict(cfg=config.dot_general_make(None, 8)),
      dict(cfg=config.dot_general_make(8, None)),
      dict(
          cfg=fqt_param_dict(s=10, use_fwd_quant=True)["cfg"],
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      dict(
          cfg=fqt_param_dict(
              s=10,
              use_fwd_quant=True,
              dlhs_local_aqt=config.LocalAqt(2),
          )["cfg"],
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      dict(
          cfg=config.dot_general_make(2, 2),
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      *[fqt_param_dict(s, use_fwd_quant=False) for s in range(10)],
      *[fqt_param_dict(s, use_fwd_quant=True) for s in range(10)],
  ])
  def test_dot_general(
      self,
      cfg: config.DotGeneral,
      lhs_maxval=10.0,
      rhs_maxval=20.0,
      gra_maxval=30.0,
      dims=(((1,), (0,)), ((), ())),  # classical matmul
      # lhs_shape=(1, 1),
      # rhs_shape=(1, 2),
      # gra_shape=(1, 2),  # has to be the shape of the output
      # lhs_shape=(1, 2),
      # rhs_shape=(2, 3),
      # gra_shape=(1, 3),  # has to be the shape of the output
      lhs_shape=(10, 20),
      rhs_shape=(20, 30),
      gra_shape=(10, 30),  # has to be the shape of the output
      seed=0,
      dtype=jnp.float32,
      clip_gradient=False,
  ):
    readonly_cfg = cfg
    del cfg

    def modify_cfg(
        *,
        use_fake_quant=False,
        use_fwd_quant=None,
        fwd_lhs_tricky_clip_and_round=False,
        local_aqt=None,
        clip_gradient=False,
    ):
      cfg = copy.deepcopy(readonly_cfg)
      if fwd_lhs_tricky_clip_and_round:
        # Tricky means that we have zero gradient on x < 0
        class TrickyNumerics:

          def abs_val_mapped_to(self) -> jnp.ndarray:
            return jnp.array(1.0)

          def fwd(self, x, context):
            del context
            return jax.lax.round(x, jax.lax.RoundingMethod.TO_NEAREST_EVEN)
            # return x

          def vjp_fwd(self, x, context):
            res = (x,)
            return self.fwd(x, context), res

          def vjp_bwd(self, res, grad):
            (x,) = res
            ret = grad * (x >= 0)
            return (ret, None)

        cfg.fwd.lhs.numerics = TrickyNumerics()
        # Needed because int8 casting would do additional clip and round.
        cfg.fwd.dg_in_dtype = None

      # Setting po2_scale is ensuring that fake_quant and full dot_general
      # have the same numerics when scales are power of two (po2).
      # We are passing dims to config so that we can reuse it in fake_quant.
      # Power-of-2 scales allow FQ and AQT to be exactly the same.
      cfg.fwd.lhs.po2_scale = True
      cfg.fwd.rhs.po2_scale = True

      cfg.dlhs.lhs.po2_scale = True
      cfg.dlhs.rhs.po2_scale = True

      cfg.drhs.lhs.po2_scale = True
      cfg.drhs.rhs.po2_scale = True

      cfg.fwd.lhs.use_fake_quant = use_fake_quant
      cfg.fwd.rhs.use_fake_quant = use_fake_quant

      cfg.dlhs.lhs.use_fake_quant = use_fake_quant
      cfg.dlhs.rhs.use_fake_quant = use_fake_quant

      cfg.drhs.lhs.use_fake_quant = use_fake_quant
      cfg.drhs.rhs.use_fake_quant = use_fake_quant

      def disable_quant_types(c):
        c.dg_in_dtype = None
        c.dg_accumulator_dtype = None

      if use_fake_quant:
        disable_quant_types(cfg.fwd)
        disable_quant_types(cfg.dlhs)
        disable_quant_types(cfg.drhs)

      if use_fwd_quant is not None:
        # If we disable all rounding, we are just testing whether the scales are
        # correct. We don't even need to disable clipping and we are testing
        # that the scales are not too large.
        def disable_quant(c):
          disable_quant_types(c)
          if isinstance(c.lhs.numerics, int_numerics.IntNumerics):
            c.lhs.numerics = c.lhs.numerics.replace(round=False)
          if isinstance(c.rhs.numerics, int_numerics.IntNumerics):
            c.rhs.numerics = c.rhs.numerics.replace(round=False)
          # c.lhs.numerics.clip = False
          # c.rhs.numerics.clip = False

        disable_quant(cfg.fwd)
        disable_quant(cfg.dlhs)
        disable_quant(cfg.drhs)
        lhs_quant = not isinstance(cfg.fwd.lhs.numerics, config.NoNumerics)
        rhs_quant = not isinstance(cfg.fwd.rhs.numerics, config.NoNumerics)
        if lhs_quant:
          cfg.drhs.rhs.use_fwd_quant = use_fwd_quant
        if rhs_quant:
          cfg.dlhs.rhs.use_fwd_quant = use_fwd_quant
      if local_aqt is not None:
        # Currently we are not supporting local_aqt in fwd pass
        # cfg.fwd.local_aqt = local_aqt
        cfg.dlhs.local_aqt = local_aqt
        cfg.drhs.local_aqt = local_aqt

      # When using abs-max scaling, this should be a no-op.
      if isinstance(cfg.fwd.lhs.numerics, int_numerics.IntNumerics):
        cfg.fwd.lhs.numerics = cfg.fwd.lhs.numerics.replace(
            clip_gradient=clip_gradient
        )
      if isinstance(cfg.fwd.rhs.numerics, int_numerics.IntNumerics):
        cfg.fwd.rhs.numerics = cfg.fwd.rhs.numerics.replace(
            clip_gradient=clip_gradient
        )

      return cfg

    # test dot_general
    lhs = rand_unif(lhs_shape, lhs_maxval, seed, dtype)
    rhs = rand_unif(rhs_shape, rhs_maxval, seed + 1, dtype)
    gra = rand_unif(gra_shape, gra_maxval, seed + 2, dtype)

    test_context = aqt.Context(key=jax.random.PRNGKey(4), train_step=None)

    def aqt_dg_full(
        use_fake_quant,
        use_fwd_quant=None,
        fwd_lhs_tricky_clip_and_round=False,
        local_aqt=None,
    ):
      cfg = modify_cfg(
          use_fake_quant=use_fake_quant,
          use_fwd_quant=use_fwd_quant,
          fwd_lhs_tricky_clip_and_round=fwd_lhs_tricky_clip_and_round,
          local_aqt=local_aqt,
          clip_gradient=clip_gradient,
      )
      dg = aqt.make_dot_general(cfg)
      return lambda lhs, rhs: dg(lhs, rhs, dims, context=test_context)

    def aqt_dg_raw(use_fake_quant):
      cfg = modify_cfg(use_fake_quant=use_fake_quant).fwd
      dg_raw = aqt._make_dot_general_raw(cfg)
      return lambda lhs, rhs: dg_raw(lhs, rhs, dims, test_context)[0]

    # Test that with backprop correctly composes 3 functions.
    # We need to test shape calculations and the returned values.
    # For the former we have multiple shapes,
    # for the latter we add some constant and test it on return.
    def lax_dg(lhs, rhs):
      return jax.lax.dot_general(lhs, rhs, dims)

    def lax_dg_248(lhs, rhs):
      def dg_mul(delta):
        def dg(
            lhs,
            rhs,
            dimension_numbers,
            context,
        ):
          if isinstance(rhs, aqt.MultiTensor):
            rhs = rhs.x

          ret = jax.lax.dot_general(lhs, rhs, dimension_numbers)
          ret *= delta

          def res(v):
            return aqt.TensorRes(
                mt=aqt.MultiTensor(
                    x=v,
                    qx=None,
                ),
                quant_grad=None,
            )

          res = aqt.DotGeneralRes(
              context_bwd=context,
              lhs=res(lhs),
              rhs=res(rhs),
          )
          return ret, res

        return dg

      m1 = dg_mul(2.0)
      m2 = dg_mul(4.0)
      m3 = dg_mul(8.0)
      return aqt._dot_general_raw_attach_gradient(m1, m2, m3)(
          lhs, rhs, dims, context=aqt.Context(key=None, train_step=None)
      )

    test_jaxpr_dtype(
        lambda: aqt_dg_full(False)(lhs, rhs),
        [modify_cfg().fwd],
        lhs.dtype,
    )
    test_jaxpr_dtype(
        lambda: jax.vjp(aqt_dg_full(False), lhs, rhs),
        [modify_cfg().fwd],
        lhs.dtype,
    )
    _, backprop = jax.vjp(aqt_dg_full(False), lhs, rhs)
    test_jaxpr_dtype(
        lambda: backprop(gra),
        [modify_cfg().dlhs, modify_cfg().drhs],
        gra.dtype,
    )

    # Test exact equalities.
    def check(dgs):
      good_lr = None
      good_gl = None
      good_gr = None
      for name, dg, options in dgs:
        # Default settings:
        if "test_gradient" not in options:
          options["test_gradient"] = True
        if "mult" not in options:
          options["mult"] = (1.0, 1.0, 1.0)
        if "check_fwd_lhs_tricky_clip_and_round" not in options:
          options["check_fwd_lhs_tricky_clip_and_round"] = False

        lrf = dg(lhs, rhs)
        lr, backprop = jax.vjp(dg, lhs, rhs)

        gl, gr = backprop(gra) if options["test_gradient"] else (None, None)

        if options["check_fwd_lhs_tricky_clip_and_round"]:
          # Test that all the expected were zero anyway
          where_zero_gradient_expected = lhs < 0
          assert (gl[where_zero_gradient_expected] == 0.0).all()

        good_lr = lr if good_lr is None else good_lr
        good_gl = gl if good_gl is None else good_gl
        good_gr = gr if good_gr is None else good_gr

        test_eq(f"{name}: lr vs lrf", lr, lrf)

        lr_mult, gl_mult, gr_mult = options["mult"]
        test_eq(f"{name}: lr", good_lr, lr / lr_mult)  # forward pass
        if options["test_gradient"]:
          test_eq(f"{name}: gl", good_gl, gl / gl_mult)  # backward pass  # pytype: disable=unsupported-operands
          test_eq(f"{name}: gr", good_gr, gr / gr_mult)  # backward pass  # pytype: disable=unsupported-operands

    check([
        ("default    ", aqt_dg_full(False), dict()),
        ("FQ         ", aqt_dg_full(True), dict()),
        ("raw fwd    ", aqt_dg_raw(False), dict(test_gradient=False)),
        ("raw fwd FQ ", aqt_dg_raw(True), dict(test_gradient=False)),
    ])

    check([
        ("fwd_quant=T", aqt_dg_full(False, use_fwd_quant=False), dict()),
        ("fwd_quant=F", aqt_dg_full(False, use_fwd_quant=True), dict()),
    ])

    check([
        (
            "default    ",
            aqt_dg_full(False, local_aqt=config.LocalAqt(2)),
            dict(),
        ),
        (
            "default    ",
            aqt_dg_full(True, local_aqt=config.LocalAqt(2)),
            dict(),
        ),
    ])

    check([
        ("lax_dg    ", lax_dg, dict()),
        ("lax_dg_248", lax_dg_248, dict(mult=(2.0, 4.0, 8.0))),
    ])

    if isinstance(readonly_cfg.fwd.lhs.numerics, int_numerics.IntNumerics):
      check([
          (
              "check_fwd_lhs_tricky_clip_and_round",
              aqt_dg_full(False, fwd_lhs_tricky_clip_and_round=True),
              dict(check_fwd_lhs_tricky_clip_and_round=True),
          ),
      ])

  def test_dynamic_context(self):
    @jax.jit
    def f(lhs, rhs, context):
      cfg = config.dot_general_make()
      dg = aqt.make_dot_general(cfg)
      return dg(lhs, rhs, (((0,), (0,)), ((), ())), context=context)

    lhs, rhs = jnp.array([3.0, 4.0]), jnp.array([4.0, 5.0])
    context = aqt.Context(
        key=jax.random.PRNGKey(4), train_step=None
    )  # xkcd.com/221
    jax.value_and_grad(f)(lhs, rhs, context)

  def test_hardware_int8(self, seed=0):
    cfg = config.dot_general_raw_make(8, 8)

    def dg(lhs, rhs):
      ret, _ = aqt._make_dot_general_raw(cfg)(
          lhs,
          rhs,
          (((1,), (0,)), ((), ())),
          aqt.Context(key=None, train_step=None),
      )
      return ret

    lhs = rand_unif((10, 20), 1.0, seed)
    rhs = rand_unif((20, 30), 1.0, seed + 1)
    test_jaxpr_dtype(lambda: dg(lhs, rhs), [cfg], lhs.dtype)
    assert cfg.dg_in_dtype == jnp.int8

  @parameterized.parameters([
      (1, 1),
      (1, 2),
      (2, 1),
      (2, 2),
      (8, 8),
      (None, 8),
      (8, None),
      (None, None),
  ])
  def test_conv_general_dilated(
      self,
      lhs_bits,
      rhs_bits,
      lhs_maxval=10.0,
      rhs_maxval=20.0,
      seed=0,
  ):
    cfg = config.conv_general_dilated_make(2, lhs_bits, rhs_bits)

    if cfg.lhs:
      # Power-of-2 scales allow FQ and AQT to be exactly the same.
      cfg.lhs.po2_scale = True
    if cfg.rhs:
      cfg.rhs.po2_scale = True

    batch_n = 10
    contr_n = 20
    feature_n = 30
    lhs = rand_unif((batch_n, 4, 5, contr_n), lhs_maxval, seed)
    rhs = rand_unif((3, 3, contr_n, feature_n), rhs_maxval, seed + 1)

    lax_conv = jax.lax.conv_general_dilated
    aqt_conv = aqt.make_conv_general_dilated(cfg)
    kwargs = {
        "window_strides": (1, 1),
        "padding": "SAME",
        "dimension_numbers": fl._conv_dimension_numbers(lhs.shape),
    }
    lhs_fq = aqt.make_fake_quant(cfg.lhs)(
        lhs, aqt.Context(key=None, train_step=None)
    )
    rhs_fq = aqt.make_fake_quant(cfg.rhs)(
        rhs, aqt.Context(key=None, train_step=None)
    )
    prod_fq = lax_conv(lhs_fq, rhs_fq, **kwargs)
    prod_aqt = aqt_conv(lhs, rhs, **kwargs)
    assert (prod_aqt == prod_fq).all()

  @parameterized.parameters([
      dict(
          shard_count=2,
          lhs=[1270.0, 10.0, 1270000.0, 10000.0],
          expected_product=1281280.0,
      ),
      dict(
          shard_count=1,
          lhs=[1270.0, 10.0, 1270000.0, 10000.0],
          expected_product=1280000.0,
      ),
  ])
  def test_local_aqt(self, shard_count, lhs, expected_product):
    # create a config that quantizes both forward and backward passes to int8
    # set the number of shards (local aqt) to 2
    cfg = config.fully_quantized(
        fwd_bits=8,
        bwd_bits=8,
        use_stochastic_rounding=False,
        drhs_local_aqt=config.LocalAqt(shard_count),
    )
    cfg.fwd.lhs.numerics = cfg.fwd.lhs.numerics.replace(preserve_max_val=True)
    cfg.fwd.rhs.numerics = cfg.fwd.rhs.numerics.replace(preserve_max_val=True)
    cfg.drhs.lhs.numerics = cfg.drhs.lhs.numerics.replace(preserve_max_val=True)
    cfg.drhs.rhs.numerics = cfg.drhs.rhs.numerics.replace(preserve_max_val=True)
    dg = lambda lhs, rhs: aqt.make_dot_general(cfg)(
        lhs,
        rhs,
        dimension_numbers=(((), ()), ((), ())),
        context=aqt.Context(key=None, train_step=None),
    )
    lhs = jnp.array(lhs)
    rhs = jnp.array([1.0])
    output, bprop = jax.vjp(dg, lhs, rhs)
    _, drhs = bprop(jnp.ones_like(output))
    assert drhs == expected_product

  def test_mnist_training(self):
    batch_size = 4
    dataset_size = 8
    target_loss = {
        "cpu": [
            3.981118679046630859375000000000,  # rome, milan
            3.981118917465209960937500000000,  # skylake
        ],
        "TPU v2": [3.991446971893310546875000000000],
        "TPU v3": [3.991446971893310546875000000000],
        "TPU v4": [3.992439270019531250000000000000],
        "TPU v5 lite": [3.991421222686767578125000000000],
    }

    class TrainConfig:
      learning_rate = 0.1
      momentum = 0.9

    def make_aqt_cfg(freeze_rhs: bool, use_frozen: bool | None):
      aqt_cfg = aqt_flax.config_v4(
          fwd_bits=8,
          dlhs_bits=8,
          drhs_bits=8,
          freeze_lhs=False,  # freeze lhs
          freeze_rhs=freeze_rhs,  # freeze rhs
          use_frozen=use_frozen,  # update sowed values
      )
      # below 3 lines are differences between config_v4/v3 and fully_quantized
      config.set_stochastic_rounding(aqt_cfg, True, True, "jax.uniform")
      aqt_cfg.dlhs.rhs.use_fwd_quant = True
      aqt_cfg.drhs.rhs.use_fwd_quant = True
      return aqt_cfg

    def forward(model, freeze_rhs: bool, use_frozen: bool | None):
      aqt_cfg = make_aqt_cfg(freeze_rhs, use_frozen)
      logits, updated_model = aqt_mnist.CNN(False, aqt_cfg).apply(
          model,
          train_ds["image"],
          rngs={"params": jax.random.PRNGKey(0)},
          mutable=True,
      )
      return logits, updated_model

    train_cfg = TrainConfig()
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    # stage 1: regular training
    aqt_cfg = make_aqt_cfg(freeze_rhs=False, use_frozen=None)
    state = aqt_mnist.create_train_state(init_rng, train_cfg, aqt_cfg)
    rng, ds_rng = jax.random.split(rng)
    train_ds = {
        "image": jax.random.uniform(
            key=ds_rng, shape=(dataset_size, 28, 28, 1)
        ),
        "label": jax.random.randint(
            key=ds_rng, shape=(dataset_size,), minval=0, maxval=10
        ),
    }
    _, input_rng = jax.random.split(rng)
    state, train_loss, _ = aqt_mnist.train_epoch(
        state, train_ds, batch_size, input_rng
    )
    assert train_loss in target_loss[jax.devices()[0].device_kind]

    # Run forward once to get the training loss
    logits_stage1, _ = forward(state.model, freeze_rhs=False, use_frozen=None)

    # stage 2: conversion / freeze weights
    logits_stage2, freezed_model = forward(
        state.model, freeze_rhs=True, use_frozen=False
    )

    assert (logits_stage2 == logits_stage1).all()
    expected_aqt_collection = {
        "Dense_0": {
            "AqtDotGeneral_0": {
                "rhs": {"frozen": jnp.int8},
                "rhs_scale": {"frozen": jnp.float32},
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "rhs": {"frozen": jnp.int8},
                "rhs_scale": {"frozen": jnp.float32},
            }
        },
    }
    aqt_collection = jax.tree_util.tree_map(lambda x: x.dtype, freezed_model)
    assert aqt_collection["aqt"] == expected_aqt_collection
    # zero out unquantized kernels
    def zero_out_params(model, layer: str):
      updated_model = copy.deepcopy(model)
      updated_model["params"][layer]["kernel"] = jnp.zeros_like(
          updated_model["params"][layer]["kernel"]
      )
      return updated_model

    freezed_model = zero_out_params(freezed_model, "Dense_0")
    freezed_model = zero_out_params(freezed_model, "Dense_1")

    # stage 3: inference mode
    logits_stage3, _ = forward(freezed_model, freeze_rhs=True, use_frozen=True)
    assert (logits_stage2 == logits_stage3).all()

    # sanity check
    sanity_model = zero_out_params(freezed_model, "Conv_0")
    logits_sanity, _ = forward(sanity_model, freeze_rhs=True, use_frozen=True)
    assert not (logits_sanity == logits_stage3).all()

    freezed_model["aqt"]["Dense_0"]["AqtDotGeneral_0"]["rhs"]["frozen"] = (
        jnp.zeros_like(
            freezed_model["aqt"]["Dense_0"]["AqtDotGeneral_0"]["rhs"]["frozen"]
        )
    )
    logits_sanity_aqt, _ = forward(
        freezed_model, freeze_rhs=True, use_frozen=True
    )
    assert not (logits_sanity_aqt == logits_stage3).all()


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for quantizing single tensor values.

The building block for quantized operations with multiple inputs
is quantizing individual inputs, possibly with functionally coupled parameters.
This module provides configurable functions for single-tensor calibration and
quantization, including static quantization allowing freeze quantization scales
and dynamic quantization with history-independent quantization scales.
quantization.
"""
import dataclasses
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from aqt.common import aqt_common
from aqt.common import aqt_config
from aqt.common import emulated_floating_points
from aqt.common import emulation_utils
import tensorflow.compat.v1 as tf
import tensorflow.compat.v1.tpu as tpu_ops


def _emulated_fp(
    t: tf.Tensor,
    mantissa_bits: tf.Tensor,
    min_exp: tf.Tensor,
    max_exp: tf.Tensor,
) -> tf.Tensor:
  """Emulates enumerics in bfloat16 or float32 given the FPMetadata.

  This is based on the implementation in:
  google3/platforms/deepsea/ffds/reduced_precision/emulated_floating_points.py
  However a separate implementation is needed to incorporate the metadata as
  tensors in the TF graph.

  Args:
    t: a tensor whose dtype is either tf.bfloat16 or tf.float32.
    mantissa_bits: The mantissa bits in the emulated format.
    min_exp: the allowed minimum exponents in the emulated format.
    max_exp: the allowed maximum exponents in the emulated format.

  Returns:
    a same dtype tensor that emulates floating points.
  """
  assert t.dtype == tf.float32
  output_type = t.dtype

  with tf.name_scope('emulated_fp'):
    v = emulated_floating_points.handle_mantissa(
        t,
        mantissa_bits=mantissa_bits,
        min_exp=min_exp,
        rounding_mode=emulation_utils.ROUND_TO_NEAREST_EVEN)
    v = emulated_floating_points.static_handle_exponent(
        v,
        min_exp=min_exp - mantissa_bits,
        max_exp=max_exp,
        mantissa_bits=mantissa_bits)
    return tf.cast(v, output_type)

# For compatibility with different frameworks, such as Babelfish, we allow
# custom variable creation lambdas. Make sure they're not trainable.
#
# GetVariable(name, shape, dtype, const_initial_value)
GetVariable = Callable[[str, Iterable[int], tf.dtypes.DType, Any], tf.Variable]


# Note type(default_get_variable) == GetVariable.
def default_get_variable(name: str, shape: Iterable[int],
                         dtype: tf.dtypes.DType, init: Any) -> tf.Variable:
  """Creates a non-trainable resource variable with tf.get_variable."""
  initializer = lambda shape, dtype: tf.constant(init, shape=shape, dtype=dtype)
  return tf.get_variable(
      name=name,
      shape=shape,
      dtype=dtype,
      trainable=False,
      initializer=initializer,
      use_resource=True)


_DivideFn = Callable[[tf.Tensor, tf.Tensor], tf.Tensor]


def _lp_dev(
    sum_of_lp_vals: tf.Tensor,  #
    sum_of_ones: tf.Tensor,
    lp_order: int,
    divide_fn: _DivideFn = tf.math.divide_no_nan,
) -> tf.Tensor:
  """Computes lp deviation."""
  if lp_order == 2:
    # sqrt() is numerically more accurate
    return tf.sqrt(divide_fn(sum_of_lp_vals, sum_of_ones))
  else:
    # TODO(b/205769820): Make sure if the output of pow op below is
    # numerically valid
    return divide_fn(sum_of_lp_vals, sum_of_ones) ** (1.0 / lp_order)


def _reduce_fn(
    stats_config: aqt_config.StatsConfig,
    s: tf.Tensor,
    weight: Optional[tf.Tensor],
    reduce_max=False,
) -> tf.Tensor:
  """Reduce function to calculate statistics."""
  if weight is not None:
    if reduce_max:
      # A maximum is a maximum, regardless of its nonzero weight.
      s = tf.where_v2(weight > 0, s, 0)
    else:
      s = s * weight
  reduce_fn = tf.math.reduce_max if reduce_max else tf.math.reduce_sum
  s = reduce_fn(s, axis=list(stats_config.share_stats_axes), keepdims=True)
  if stats_config.tpu_cross_replica_sum:
    raise NotImplementedError(
        'support for tpu_cross_replica_sum=True is not implemented'
    )
    s = tpu_ops.cross_replica_sum(s)  # pylint:disable=unreachable
  return s


def _sum_of_ones(
    stats_config: aqt_config.StatsConfig,
    x: tf.Tensor,
    weight: Optional[tf.Tensor],
) -> tf.Tensor:
  """Computes the number of entries along non-shared stats axes (maybe filter zeros)."""
  # Layers such as ReLU emit zeros often. In such cases, we can model
  # the non-sparse distribution of weights separately, resulting in
  # unbiased estimation of non-sparse mean l1 and lp.
  # This clips away less of the distribution of inputs.
  if stats_config.filter_zeros:
    ones = tf.cast(tf.math.not_equal(x, 0), dtype=tf.float32)
  else:
    ones = tf.ones_like(x)
  return _reduce_fn(stats_config, ones, weight)


def _max_of_abs_vals(
    stats_config: aqt_config.StatsConfig,
    x: tf.Tensor,
    weight: Optional[tf.Tensor],
) -> tf.Tensor:
  return _reduce_fn(stats_config, tf.abs(x), weight, reduce_max=True)


_sum_of_vals = _reduce_fn


def _sum_of_l1_vals(
    stats_config: aqt_config.StatsConfig,
    x: tf.Tensor,
    weight: Optional[tf.Tensor],
) -> tf.Operation:
  return _reduce_fn(stats_config, tf.abs(x), weight)


def _sum_of_lp_vals(
    stats_config: aqt_config.StatsConfig,
    x: tf.Tensor,
    weight: Optional[tf.Tensor],
) -> tf.Operation:
  # Below, we need not excerpt the zeros from updates to sum of
  # vals, l1, and lp, because zeros do not affect those aggregates.

  # Avoid unnecessary tf.abs for even powers.
  px = x if stats_config.lp_order % 2 == 0 else tf.abs(x)

  return _reduce_fn(stats_config, px**stats_config.lp_order, weight)


def get_stats_shape(
    share_stats_axes: Iterable[int], data_shape: Iterable[Optional[int]]
) -> List[Optional[int]]:
  """Returns the shape of the statistics.

  Replaces the dimensions in the data shape with ones where we share statistics.

  Args:
    share_stats_axes: axes where statistics are shared.
    data_shape: shape of a tensor.

  Returns:
    the shape of statistics.
  """
  stats_shape = list(data_shape)
  for axis in share_stats_axes:
    stats_shape[axis] = 1
  return stats_shape


def _init_dynamic_stats(
    stats_config: aqt_config.StatsConfig,
    x: tf.Tensor,
    init_value: float = 0.0,
) -> tf.Tensor:
  """Initializes a dynamic statistical tensor."""
  # assume x is of dynamic shape and we want to have a constant tensor with the
  # shape of x except for shared statistics axes where dimensions are ones.
  rank = len(x.shape.as_list())
  indices = []
  for i in range(rank):
    if i in stats_config.share_stats_axes:
      indices.append(slice(0, 1))
    else:
      indices.append(slice(None))
  ones = tf.ones_like(x[indices], dtype=x.dtype)
  return ones * init_value


def _bound(
    calibration_config: aqt_config.CalibrationConfig,
    lp_order: int,
    init_bound: tf.Tensor,
    sum_of_ones: Optional[tf.Tensor],
    max_of_abs_vals: Optional[tf.Tensor],
    sum_of_l1_vals: Optional[tf.Tensor],
    sum_of_lp_vals: Optional[tf.Tensor],
    divide_fn: _DivideFn,
) -> tf.Tensor:
  """Computes the upper bound."""
  bound = init_bound + calibration_config.const_bound_coeff
  if calibration_config.l1_dev_coeff:
    l1_dev = divide_fn(sum_of_l1_vals, sum_of_ones)
    bound += calibration_config.l1_dev_coeff * l1_dev
  if calibration_config.lp_dev_coeff:
    lp_dev = _lp_dev(sum_of_lp_vals, sum_of_ones, lp_order, divide_fn)
    bound += calibration_config.lp_dev_coeff * lp_dev
  if calibration_config.max_dev_coeff:
    max_dev = max_of_abs_vals
    bound += calibration_config.max_dev_coeff * max_dev
  return bound


def _dynamic_bound(
    config: aqt_config.StatsConfig,
    calibration_config: aqt_config.CalibrationConfig,
    x: tf.Tensor,
    weight: Optional[tf.Tensor],
) -> tf.Tensor:
  """Compute the upper bound on input tensor values dynamically."""
  config.validate(x.shape.as_list(), dynamic=True)
  init_bound = _init_dynamic_stats(config, x, init_value=0.0)
  divide_fn = tf.math.divide_no_nan if config.safe_divide else tf.divide
  sum_of_ones = max_of_abs_vals = sum_of_l1_vals = sum_of_lp_vals = None
  if any([
      calibration_config.l1_dev_coeff,
      calibration_config.lp_dev_coeff,
      calibration_config.max_dev_coeff,
  ]):
    sum_of_ones = _sum_of_ones(config, x, weight)
  if calibration_config.max_dev_coeff:
    max_of_abs_vals = _max_of_abs_vals(config, x, weight)
  if calibration_config.l1_dev_coeff:
    sum_of_l1_vals = _sum_of_l1_vals(config, x, weight)
  if calibration_config.lp_dev_coeff:
    sum_of_lp_vals = _sum_of_lp_vals(config, x, weight)
  return _bound(
      calibration_config,
      config.lp_order,
      init_bound,
      sum_of_ones,
      max_of_abs_vals,
      sum_of_l1_vals,
      sum_of_lp_vals,
      divide_fn,
  )


class Stats:
  """Manages efficient gathering of running statistics."""

  def __init__(
      self,  #
      *,
      data_shape: Iterable[Optional[int]],
      config: aqt_config.StatsConfig,
      get_variable: GetVariable):
    self._data_shape = list(data_shape)
    config.validate(self._data_shape)
    if config.lp_order > 30:
      raise NotImplementedError('For higher norms we should add stabilization.')
    self._config = config
    self._ema_update_count = self._config.ema_update_count

    self.stats_shape = get_stats_shape(
        self._config.share_stats_axes,
        self._data_shape,
    )

    self.divide = (tf.math.divide_no_nan if self._config.safe_divide
                   else tf.math.divide)

    def mk_var(name, init_val):
      return get_variable(name, self.stats_shape, tf.float32, init_val)

    self._sum_of_ones = mk_var('sum_of_ones', self._config.update_count_prior)
    self._sum_of_vals = mk_var(
        'sum_of_vals',
        self._config.mean_prior * self._config.update_count_prior)
    self._max_of_abs_vals = mk_var('max_of_abs_vals',
                                   self._config.max_dev_prior)
    self._sum_of_l1_vals = mk_var(
        'sum_of_l1_vals',
        self._config.l1_dev_prior * self._config.update_count_prior)
    self._sum_of_lp_vals = mk_var(
        'sum_of_lp_vals', self._config.lp_dev_prior**self._config.lp_order *
        self._config.update_count_prior)

  def update(self, x: tf.Tensor, weight: Optional[tf.Tensor]) -> tf.Operation:
    """Updates internal statistics with the new observation."""
    aqt_common.check_shapes_conformal(x.shape.as_list(), self._data_shape)
    if weight is not None and len(x.shape) != len(weight.shape):
      raise ValueError(
          f'expected rank(x)={len(x.shape)} == rank(weight)={len(weight.shape)}'
      )

    def update_var(var, update_fn):
      s = update_fn(self._config, x, weight)
      rate = 1.0 / self._ema_update_count
      return var.assign((1.0 - rate) * var.read_value() + rate * s)

    return tf.group([
        update_var(self._sum_of_ones, _sum_of_ones),
        update_var(self._sum_of_vals, _sum_of_vals),
        update_var(self._max_of_abs_vals, _max_of_abs_vals),
        update_var(self._sum_of_l1_vals, _sum_of_l1_vals),
        update_var(self._sum_of_lp_vals, _sum_of_lp_vals),
    ])

  def mean(self) -> tf.Tensor:
    return self.divide(self._sum_of_vals.read_value(),
                       self._sum_of_ones.read_value())

  def l1_dev(self) -> tf.Tensor:
    return self.divide(self._sum_of_l1_vals.read_value(),
                       self._sum_of_ones.read_value())

  def lp_dev(self) -> tf.Tensor:
    return _lp_dev(
        self._sum_of_lp_vals.read_value(),
        self._sum_of_ones.read_value(),
        self._config.lp_order,
        self.divide,
    )

  def max_dev(self) -> tf.Tensor:
    return self._max_of_abs_vals

  def bound(  #
      self, calibration_config: aqt_config.CalibrationConfig) -> tf.Tensor:
    """Compute the upper bound on input tensor values, broadcastable to input."""
    return _bound(
        calibration_config,
        self._config.lp_order,
        tf.zeros(self.stats_shape, dtype=tf.float32),
        self._sum_of_ones.read_value(),
        self._max_of_abs_vals,
        self._sum_of_l1_vals.read_value(),
        self._sum_of_lp_vals.read_value(),
        self.divide,
    )

  def calibration_variables(self) -> Dict[str, tf.Variable]:
    """Returns variables used for self.bound()."""
    return {
        'sum_of_ones': self._sum_of_ones,  #
        'sum_of_vals': self._sum_of_vals,
        'max_of_abs_vals': self._max_of_abs_vals,
        'sum_of_l1_vals': self._sum_of_l1_vals,
        'sum_of_lp_vals': self._sum_of_lp_vals,
    }


def is_config_active(config: aqt_config.AqtTensorConfig,
                     event_count: tf.Tensor) -> tf.Tensor:
  """Return whether the configuration is active at event_count."""
  config.validate()
  should_q = tf.constant(True)
  if config.begin_at_event is not None:
    should_q &= config.begin_at_event <= event_count
  if config.end_at_event is not None:
    should_q &= event_count < config.end_at_event
  return should_q


def _should_update_scale(
    config: aqt_config.AqtTensorConfig,  #
    prev_event_count: tf.Tensor,
    new_event_count: tf.Tensor) -> tf.Tensor:
  """Returns if scale should be updated for the config and event count."""
  if isinstance(config.quant_config, aqt_config.FloatConfig):
    return tf.constant(False)

  if not config.freeze_scale_at_begin:
    return tf.constant(True)

  # The first time a config is active, even if we freeze scale, we should
  # update the scale.
  was_previously_inactive = ~is_config_active(config, prev_event_count)

  # We rely on tf.int64.min being an illegal event count value, so that
  # even if is_config_active(config, tf.int64.min), we still update scale.
  # This could happen if the very first update happens for a config which
  # has freeze_scale_at_begin and begin=None.
  assert_op = tf.debugging.assert_greater(
      new_event_count, tf.int64.min, message='event_count cannot be int64.min')
  with tf.control_dependencies([assert_op]):
    first_event = tf.math.equal(prev_event_count, tf.int64.min)

  return was_previously_inactive | first_event


class TensorQuantizerBase:
  """Maintains state associated with the quantization of an input tensor.

  A TensorQuantizerBase provides common functionalities for both static and
  dynamic quantization, recording the last time of the most recent update to
  this `TensorQuantizerBase` class to be able to switch between `tensor_configs`
  in a given `AqtScheduleConfig`.

  TensorQuantizer assumes the supplied `event_count` strictly monotonically
  increases and starts out strictly greater than `tf.int64.min`.

  Uses the `get_variable` provided in the constructor to create variables.

  Attributes:
    data_shape: the shape of the tensor this quantizes. These dimensions may be
      `None`, but then they must correspond to shared stats axes in
      `config.stats_config` when quantization is static. In this case, use of
      quanitzed variable is disallowed as the shapes of the quantization scales
      are not fully defined.
    config: A training quantization schedule.
  """

  def __init__(
      self,  #
      data_shape: Iterable[Optional[int]],
      config: aqt_config.AqtScheduleConfig,
      get_variable: GetVariable = default_get_variable,
      name: str = 'tensor_quantizer_base',
  ):
    self.data_shape = list(data_shape)
    config.fill_gaps_with_float_config()
    config.validate(self.data_shape)
    self.config = config

    with tf.variable_scope(name):
      # This variable maintains the most recent event count at which this
      # TensorQuantizer was updated. This determines which quantization config
      # from our schedule `self.config` is active, which is required because:
      # (1) at inference time, this determines the quantization config to serve.
      # (2) at training time, the first time a config with a frozen scale is
      # active, we must update the scale, a calculation requiring this variable.
      self._last_update = get_variable('last_update', [], tf.int64,
                                       tf.int64.min)

  def tracked_variables(self) -> Dict[str, tf.Variable]:
    """Returns variables used to track updates."""
    return {
        'last_update': self._last_update,
    }

  def calibration_variables(self) -> Dict[str, tf.Variable]:
    """Returns scale and stats variables used to calibrate tensors."""
    raise NotImplementedError

  def _config_case(
      self,  #
      case_fn: Callable[[aqt_config.AqtTensorConfig], tf.Tensor],
      event_count: tf.Tensor,
  ) -> tf.Tensor:
    """Switches over configs, applying case_fn to active one at event_count."""
    assert self.config.tensor_configs, 'There must be at least one config.'

    def make_case(config):
      pred = is_config_active(config, event_count)
      return pred, lambda: case_fn(config)

    cases = [make_case(c) for c in self.config.tensor_configs]
    return tf.case(cases, exclusive=True)

  def _quantization_params(self, train):
    """Returns parameters for AQT quantization routine."""

    @dataclasses.dataclass
    class QParams:
      """Dataclass for quantization parameters."""
      # Tensors used for quantization
      should_quantize: tf.Tensor = tf.constant(0.0)
      clip_bound: tf.Tensor = tf.constant(0.0)
      shift_before: tf.Tensor = tf.constant(0.0)
      shift_after: tf.Tensor = tf.constant(0.0)

      # Tensors used for floating point emulation.
      should_use_small_float: tf.Tensor = tf.constant(0)
      mantissa_bits: tf.Tensor = tf.constant(0)
      min_exp: tf.Tensor = tf.constant(0)
      max_exp: tf.Tensor = tf.constant(0)

      def __iter__(self):
        return (getattr(self, field.name) for field in dataclasses.fields(self))

    def qparams(config: aqt_config.AqtTensorConfig,
                check_active: bool = True) -> QParams:
      """Returns quant parameters for fixed AQT config."""

      params = QParams()
      config_active = is_config_active(config, self._last_update.read_value())
      config_active = not check_active or config_active

      if isinstance(config.quant_config, aqt_config.FloatConfig):
        params.clip_bound = tf.where_v2(config_active, float('inf'), 0.0)
      elif isinstance(config.quant_config, aqt_config.IntQuantConfig):
        config_active = tf.cast(config_active, tf.float32)
        params.should_quantize += config_active

        # TODO(vladf): some serving environments, such as adbrain,
        # automatically rewrite constants from f32 to bf16, which makes the
        # epsilon used in the function below invalid, so that
        # safe_clip_bound == clip_bound (incorrectly). We should solve for
        # that through some configuration.
        params.clip_bound += config_active * aqt_common.safe_clip_bound(
            config.quant_config)
        if config.quant_config.preserve_zero:
          params.shift_before += config_active * 0.5
        else:
          params.shift_after += config_active * 0.5
      elif isinstance(config.quant_config, aqt_config.SmallFloatConfig):
        # Note: In the case of small floats, we can just use get_clip_bound
        # as opposed to safe_clip_bound since we are mapping the clip bound
        # directly to the highest float that can be represented, while in int
        # world we are mapping biggest value onto 127.4999 to get a slightly
        # wider and less biased range.
        config_active = tf.cast(config_active, tf.int32)
        params.should_use_small_float += config_active

        params.clip_bound += (
            tf.cast(config_active, tf.float32) *
            aqt_common.get_clip_bound(config.quant_config))

        params.mantissa_bits += (
            config_active * config.quant_config.mantissa_bits)
        params.min_exp += config_active * config.quant_config.min_exp
        params.max_exp += config_active * config.quant_config.max_exp
      else:
        raise ValueError('config.quant_config must have type '
                         'FloatConfig, IntQuantConfig, or SmallFloatConfig. '
                         f'But instead got {type(config.quant_config)}.')

      return params

    if not train and self.config.inference_config_index is not None:
      params = qparams(
          self.config.tensor_configs[self.config.inference_config_index],
          check_active=False)
    elif self.config.tensor_configs:
      params = QParams(*map(tf.add_n, zip(
          *map(qparams, self.config.tensor_configs))))
    else:
      return QParams()

    params.should_quantize = tf.cast(params.should_quantize, tf.bool)
    params.should_use_small_float = tf.cast(
        params.should_use_small_float, tf.bool)
    return params

  def should_quantize(self, train: bool) -> tf.Tensor:
    """Returns the number of quant configs are active."""
    params = self._quantization_params(train)
    return params.should_quantize

  def _to_quant(
      self,
      x: tf.Tensor,
      train: bool,
      use_stochastic_rounding: bool = False,
  ) -> tf.Tensor:
    """Quantizes x with active quant config, if any, else act as identity.

    Args:
      x: tensor to quantize
      train: whether training
      use_stochastic_rounding: whether to add random numbers in [-0.5, 0.5]
        before integer quantization.

    Returns:
      quantized tensor
    """
    with tf.variable_scope('to_quant'):
      params = self._quantization_params(train)

      def _floor(t: tf.Tensor, use_stochastic_rounding: bool) -> tf.Tensor:
        if use_stochastic_rounding:
          t = t + tf.random.uniform(tf.shape(t), -0.5, 0.5, dtype=t.dtype)
        return tf.math.floor(t)

      def maybe_floor_or_small_float(y):
        # Static check for whether int and small float can coexist in schedule
        if self.config.allow_int_small_float:
          return tf.where_v2(params.should_use_small_float,
                             _emulated_fp(y, params.mantissa_bits,
                                          params.min_exp,
                                          params.max_exp),
                             tf.where_v2(params.should_quantize,
                                         tf.math.floor(y), y))
        else:
          if self.config.quantization_mode() == aqt_config.SmallFloatConfig:
            return tf.where_v2(
                params.should_use_small_float,
                _emulated_fp(y, params.mantissa_bits, params.min_exp,
                             params.max_exp), y)
          else:
            return tf.where_v2(
                params.should_quantize, _floor(y, use_stochastic_rounding), y
            )

      # Note that backprop does not depend directly on the value of _last_update
      # or any_config_active; only scales and constants need to be maintained
      # and there's no branching on ops including the input tensor x. This
      # results in significant memory reduction, see cl/415355150.
      x = tf.clip_by_value(x, -params.clip_bound, params.clip_bound)
      x += params.shift_before
      x = tf.grad_pass_through(maybe_floor_or_small_float)(x)
      x += params.shift_after

    return x

  def _clip_mask(self, x: tf.Tensor, train: bool) -> tf.Tensor:
    """Returns which entries of x are clipped in _to_quant(x)."""
    # TODO(vladf): we should consider re-using the same clip mask
    # computation in _to_quant to derive this clip mask instead of a separate
    # method.
    with tf.variable_scope('clip_mask'):
      _, clip_bound, _, _, _, _, _, _ = self._quantization_params(train)
      return tf.abs(x) > clip_bound

  def _should_scale(self, train: bool) -> tf.Tensor:
    """Returns True if any non-float quant config is active."""
    if not train and self.config.inference_config_index is not None:
      inference_config = self.config.tensor_configs[
          self.config.inference_config_index
      ]
      should_scale = tf.constant(
          not isinstance(inference_config.quant_config, aqt_config.FloatConfig)
      )
    else:
      should_scale = tf.constant(False)
      for config in self.config.tensor_configs:
        if isinstance(config.quant_config, aqt_config.FloatConfig):
          continue

        config_active = is_config_active(config, self._last_update)
        should_scale |= config_active
    return should_scale

  def _maybe_fallback_to_ones(
      self,
      should_scale: tf.Tensor,
      scale: tf.Tensor,
      inv_scale: tf.Tensor,
  ) -> Tuple[tf.Tensor, tf.Tensor]:
    """Fallback to ones if should not scale."""
    # TODO(lew): We can simplify the scopes to just 'compute scales'
    with tf.variable_scope('to_quant'):
      scale = tf.where_v2(should_scale, scale, tf.ones_like(scale))  #
    with tf.variable_scope('from_quant'):
      inv_scale = tf.where_v2(  #
          should_scale, inv_scale, tf.ones_like(inv_scale)
      )
    return scale, inv_scale


class TensorQuantizer(TensorQuantizerBase):
  """Maintains state associated with the quantization of an input tensor.

  A TensorQuantizer owns observed statistics for an input tensor, along with
  variables for the scale and event_count copy, recording the last time of the
  most recent update to this `TensorQuantizer` class.

  This class provides state-mutating methods for updating statistics for every
  observation of an input tensor and is used by AQT methods to derive
  calibration bounds.

  TensorQuantizer assumes the supplied `event_count` strictly monotonically
  increases across `TensorQuantizer.update` calls and starts out strictly
  greater than `tf.int64.min`.

  Uses the `get_variable` provided in the constructor to create variables.

  Attributes:
    data_shape: the shape of the tensor this quantizes. These dimensions may be
      `None`, but then they must correspond to shared stats axes in
      `config.stats_config`. In this case, use of quanitzed variable is
      disallowed.
    config: A training quantization schedule.
    quantized_variable: If `config.use_quantized_variable`, then the quantized
      version of the most recent `update()` tensor is saved to this member. This
      is helpful in inference settings, where users may be interested in only
      saving quantized versions of the weights to reduce storage consumption or
      avoid quantization of floating point weights at inference time.
  """

  def __init__(
      self,  #
      data_shape: Iterable[Optional[int]],
      config: aqt_config.AqtScheduleConfig,
      get_variable: GetVariable = default_get_variable,
      name: str = 'tensor_quantizer',
  ):
    super().__init__(
        data_shape=data_shape,
        config=config,
        get_variable=get_variable,
        name=name,
    )

    with tf.variable_scope(name):
      self._stats = Stats(
          data_shape=self.data_shape,  #
          config=self.config.stats_config,
          get_variable=get_variable,
      )

      # We intentionally initialize scale to zero to fail loudly if someone uses
      # a parameter such as scale without properly update()-ing it.
      self._scale = get_variable(
          'scale', self._stats.stats_shape, tf.float32, 0
      )
      # Save the inverse scale so that we don't recompute it at inference time.
      self._inv_scale = get_variable(
          'inv_scale', self._stats.stats_shape, tf.float32, 0
      )

      # Variable to save or read quantized tensors to, if the config says so.
      if self.config.use_quantized_variable:
        self.quantized_variable = get_variable(
            'quantized_variable', self.data_shape, tf.int8, 0
        )

  def tracked_variables(self) -> Dict[str, tf.Variable]:
    """Returns variables used to track updates and calibration variables."""
    variables = super().tracked_variables()
    variables.update(self.calibration_variables())
    if self.config.use_quantized_variable:
      variables['quantized_variable'] = self.quantized_variable
    return variables

  def calibration_variables(self) -> Dict[str, tf.Variable]:
    """Returns scale and stats variables used to calibrate tensors."""
    return {
        'scale': self._scale,
        'inv_scale': self._inv_scale,
        **self._stats.calibration_variables(),
    }

  def _fresh_scale(
      self, config: aqt_config.AqtTensorConfig
  ) -> Tuple[tf.Tensor, tf.Tensor]:
    """Returns new scale, inverse scale for a given config and stats, if any."""
    if isinstance(config.quant_config, aqt_config.FloatConfig):
      # We shouldn't update the scale if the given config contains FloatConfig
      # and no emulation;
      # fill with a poison value if we get into this situation.
      nan = tf.constant(float('nan'), tf.float32, self._stats.stats_shape)
      return nan, nan

    x_bound = self._stats.bound(config.calibration_config)
    clip_bound = aqt_common.get_clip_bound(config.quant_config)

    new_scale = clip_bound / x_bound
    inv_scale = x_bound / clip_bound
    return new_scale, inv_scale

  def clip_range(self) -> tf.Tensor:
    """Returns the tensor clip range or zeros if no int config is active."""

    def case_fn(config: aqt_config.AqtTensorConfig) -> tf.Tensor:
      if isinstance(config.quant_config, aqt_config.FloatConfig):
        return tf.zeros(self._stats.stats_shape)

      # We return the range derived from the inverse scale, rather than
      # from the stats themselves, to respect freezing settings and
      # report the clipping range that's actually used at all times.
      #
      # The counterfactual clipping range that would have been used
      # if we didn't freeze the scale can be re-derived from the current
      # stats values, which are updated regardless of freezing.
      clip_bound = aqt_common.get_clip_bound(config.quant_config)
      return self._inv_scale.read_value() * clip_bound

    return self._config_case(case_fn, self._last_update.read_value())

  def update(
      self,  #
      sample: Optional[tf.Tensor],
      weight: Optional[tf.Tensor],
      event_count: tf.Tensor,
  ) -> tf.Operation:
    """Op to update statistics, scale, and quantized variable.

    Args:
      sample: an observation of the tensor to quantize. If None, only update the
        state variables without updating the stats
      weight: the weight of each observation for updating statistics.
      event_count: the event count this observation corresponds to; this
        determines the scale that's active for any quantized methods referencing
        this `TensorQuantizer`.

    Returns:
      A tensorflow operation corresponding to the updates to internal variables
      for capturing this observation of `sample`.
    """
    dependencies = []
    if sample is not None:
      # update the stats if a sample is passed, else only update state
      dependencies.append(self._stats.update(sample, weight))
    with tf.control_dependencies(dependencies):

      def case_fn(config):
        return self._update_state_config(config, sample, event_count)

      return self._config_case(case_fn, event_count).op

  def _update_state_config(
      self,  #
      config: aqt_config.AqtTensorConfig,
      sample: tf.Tensor,
      event_count: tf.Tensor,
  ) -> tf.Tensor:
    """Returns tensor with dependency on updates for state variables."""
    updates = []

    # Ensure we read _last_update before we update it.
    last_update = self._last_update.read_value()
    with tf.control_dependencies([last_update]):
      should_update_scale = _should_update_scale(
          config, prev_event_count=last_update, new_event_count=event_count  #
      )
      updates.append(self._last_update.assign(event_count))

    updated_scale, updated_inv_scale = tf.cond(
        should_update_scale,  #
        lambda: self._fresh_scale(config),
        lambda: (self._scale.read_value(), self._inv_scale.read_value()),
    )
    updates.append(self._scale.assign(updated_scale))
    updates.append(self._inv_scale.assign(updated_inv_scale))

    if (
        self.config.use_quantized_variable
        and isinstance(config.quant_config, aqt_config.IntQuantConfig)
        and config.quant_config.compatible_with_int8()
    ):
      with tf.control_dependencies(updates):
        scale, _ = self._get_quant_scale(train=True)
        sample = scale * sample
        updated_quantized_variable = tf.cast(
            self._to_quant(sample, train=True), self.quantized_variable.dtype
        )
        updates.append(
            self.quantized_variable.assign(updated_quantized_variable)
        )

    with tf.control_dependencies(updates):
      return tf.constant(0)

  def _get_quant_scale(self, train: bool) -> tf.Tensor:
    """Returns scales to quantize/dequantize the active quant config, if any, else ones."""
    should_scale = self._should_scale(train)

    return self._maybe_fallback_to_ones(
        should_scale,
        self._scale,
        self._inv_scale,
    )


def validate_dynamic(config: aqt_config.AqtScheduleConfig) -> None:
  """Validates the config conforms with dynamic quantization."""
  if config.stats_config.ema_update_count != 1:
    raise aqt_config.ConfigError(
        'ema_update_count={config.stats_config.ema_update_count} must be 1 '
        'for dynamic quantization.'
    )
  if config.use_quantized_variable:
    raise aqt_config.ConfigError(
        'dynamic quantization  does not memorized the quantized variable as '
        'it is history-independent.'
    )
  for tensor_config in config.tensor_configs:
    # When a tensor config is not FloatConfig, the quantizer may use
    # non-trivial quantization scales and dynamic quantization should not use
    # scales in such cases.
    if (
        isinstance(tensor_config, aqt_config.FloatConfig)
        and tensor_config.freeze_scale_at_begin
    ):
      raise aqt_config.ConfigError(
          'Dynamic quantization should not freeze_scale_at_begin for non-float '
          'config but got {tensor_config}.'
      )


class DynamicTensorQuantizer(TensorQuantizerBase):
  """Maintains state associated with the dynamic quantization of an input tensor.

  A DynamicTensorQuantizer quantize a input tensor dynamically. It does not
  maintain any statistics, scales, or quantized variables for the forward pass.

  DynamicTensorQuantizer inherits from TensorQuantizerBase that uses the
  `get_variable` provided in the constructor to create a variable to memorize
  the most recent event_count and provides methods to switch between
  `tensor_configs` in a given `AqtScheduleConfig`

  DynamicTensorQuantizer assumes the supplied `event_count` strictly
  monotonically increases and starts out strictly greater than `tf.int64.min`.


  Attributes:
    data_shape: the shape of the tensor this quantizes. These dimensions may be
      `None`.
    config: A training quantization schedule.
  """

  def __init__(
      self,  #
      data_shape: Iterable[Optional[int]],
      config: aqt_config.AqtScheduleConfig,
      get_variable: GetVariable = default_get_variable,
      name: str = 'dynamic_tensor_quantizer',
  ):
    validate_dynamic(config)
    super().__init__(
        data_shape=data_shape,
        config=config,
        get_variable=get_variable,
        name=name,
    )

  def calibration_variables(self) -> Dict[str, tf.Variable]:
    """Returns empty dict as dynamic quantization does not store calibration variables."""
    return {}

  def dynamic_clip_range(
      self,
      sample: tf.Tensor,
      weight: Optional[tf.Tensor],
      event_count: tf.Tensor,
      train: bool,
  ) -> tf.Tensor:
    """Returns the tensor clip range or zeros if no int config is active."""
    _, inv_scale = self._get_dynamic_quant_scale(
        sample, weight, event_count, train
    )

    def case_fn(config: aqt_config.AqtTensorConfig) -> tf.Tensor:
      if isinstance(config.quant_config, aqt_config.FloatConfig):
        return tf.zeros_like(inv_scale, dtype=inv_scale.dtype)

      # We return the range derived from the inverse scale, rather than
      # from the stats themselves, to respect freezing settings and
      # report the clipping range that's actually used at all times.
      #
      # The counterfactual clipping range that would have been used
      # if we didn't freeze the scale can be re-derived from the current
      # stats values, which are updated regardless of freezing.
      clip_bound = aqt_common.get_clip_bound(config.quant_config)
      return inv_scale * clip_bound

    return self._config_case(case_fn, self._last_update.read_value())

  def _fresh_dynamic_scale(
      self,
      tensor_config: aqt_config.AqtTensorConfig,
      sample: tf.Tensor,
      weight: tf.Tensor,
  ) -> Tuple[tf.Tensor, tf.Tensor]:
    """Returns new scale, inverse scale for a given config and stats, if any."""
    if isinstance(tensor_config.quant_config, aqt_config.FloatConfig):
      # We shouldn't return the scale if the given config contains FloatConfig
      # and no emulation;
      # fill with a poison value if we get into this situation.
      nan = _init_dynamic_stats(
          self.config.stats_config, sample, init_value=float('nan'))
      return nan, nan

    x_bound = _dynamic_bound(
        self.config.stats_config,
        tensor_config.calibration_config,
        sample,
        weight,
    )

    clip_bound = aqt_common.get_clip_bound(tensor_config.quant_config)

    new_scale = clip_bound / x_bound
    inv_scale = x_bound / clip_bound
    return new_scale, inv_scale

  def _get_dynamic_quant_scale(
      self,
      sample: tf.Tensor,
      weight: Optional[tf.Tensor],
      event_count: tf.Tensor,
      train: bool,
  ) -> tf.Tensor:
    """Returns scales to quantize/dequantize the active quant config, if any, else ones."""

    # We intentionally initialize scale to zero to fail loudly if someone uses
    # a parameter such as scale without properly update()-ing it.
    zeros = _init_dynamic_stats(self.config.stats_config, sample, init_value=0)
    def case_fn(config):
      # only need to update the event_count for dynamic quantizer
      updates = []

      # Ensure we read _last_update before we update it.
      last_update = self._last_update.read_value()
      with tf.control_dependencies([last_update]):
        should_update_scale = _should_update_scale(
            config, prev_event_count=last_update, new_event_count=event_count  #
        )
        updates.append(self._last_update.assign(event_count))

      with tf.control_dependencies(updates):
        scale, inv_scale = tf.cond(
            should_update_scale,  #
            lambda: self._fresh_dynamic_scale(config, sample, weight),
            lambda: (zeros, zeros),
        )
        return scale, inv_scale

    scale, inv_scale = self._config_case(case_fn, event_count)

    # make sure the scale is updated
    with tf.control_dependencies([scale.op, inv_scale.op]):
      should_scale = self._should_scale(train)

    return self._maybe_fallback_to_ones(should_scale, scale, inv_scale)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for einsum."""

import typing
from typing import Any, Dict, Optional, Sequence, Tuple
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from aqt.common import aqt_common
from aqt.common import aqt_config
from aqt.tensorflow import aqt_einsum
from aqt.tensorflow import aqt_ops
from aqt.tensorflow import aqt_tensor
from aqt.test import aqt_test_shared_base
import numpy as np
import tensorflow.compat.v1 as tf


def _stats_config(share_stats_axes: Sequence[int]) -> aqt_config.StatsConfig:
  """Generates dynamic quantization stats configuration."""
  return aqt_config.StatsConfig(
      ema_update_count=1,
      share_stats_axes=list(share_stats_axes),
      update_count_prior=0,
      tpu_cross_replica_sum=False)


def _schedule_config(
    bits: int, const_bound_coeff: float,
    share_stats_axes: Sequence[int],
    freeze_scale_at_begin: bool = True) -> aqt_config.AqtScheduleConfig:
  """Creates schedule config with dynamic quantization."""
  iqc = aqt_config.IntQuantConfig(bits=bits)
  cc = aqt_config.CalibrationConfig(const_bound_coeff=const_bound_coeff)
  tc = aqt_config.AqtTensorConfig(
      quant_config=iqc,
      calibration_config=cc,
      freeze_scale_at_begin=freeze_scale_at_begin
      )
  sc = _stats_config(share_stats_axes)
  return aqt_config.AqtScheduleConfig(sc, [tc])


def _schedule_config_emulation(
    share_stats_axes) -> aqt_config.AqtScheduleConfig:
  """Creates schedule config for emulated precision."""
  iqc = aqt_config.SmallFloatConfig(
      exponent_bits=5,
      mantissa_bits=2,
      min_exp=-14,
      max_exp=15,
      support_inf=False,
      rounding_mode=aqt_config.RoundingMode.ROUND_TO_NEAREST_EVEN)
  # Using the max number essentially disables scaling.
  cc = aqt_config.CalibrationConfig(
      const_bound_coeff=aqt_common._get_max_number_float(
          mantissa_bits=2, max_exp=15))
  tc = aqt_config.AqtTensorConfig(
      quant_config=iqc, calibration_config=cc, freeze_scale_at_begin=True)
  sc = aqt_config.StatsConfig(
      ema_update_count=1,
      share_stats_axes=list(share_stats_axes),
      update_count_prior=0,
      tpu_cross_replica_sum=False)
  return aqt_config.AqtScheduleConfig(sc, [tc])


def _empty_config(
    share_stats_axes: Sequence[int]) -> aqt_config.AqtScheduleConfig:
  return aqt_config.AqtScheduleConfig(_stats_config(share_stats_axes), [])


def _einsum_op(
    eq: str,  #
    lhs: tf.Tensor,
    rhs: tf.Tensor,
    lhs_config: aqt_config.AqtScheduleConfig,
    rhs_config: aqt_config.AqtScheduleConfig,
    lhs_weights: Optional[tf.Tensor] = None,
    rhs_weights: Optional[tf.Tensor] = None,
    varscope_name: str = "einsum",
    train: bool = True,
    quantize_bwd: bool = False,
    lhs_bwd_config: Optional[aqt_config.AqtScheduleConfig] = None,
    rhs_bwd_config: Optional[aqt_config.AqtScheduleConfig] = None,
    dynamic_fwd_quant: bool = False,
    **einsum_kwargs,
) -> tf.Tensor:
  """Updates quantizers at event_count=0 and computes einsum."""
  with tf.variable_scope(varscope_name):
    tensor_quantizer_cls = (
        aqt_tensor.DynamicTensorQuantizer if dynamic_fwd_quant
        else aqt_tensor.TensorQuantizer
    )
    lhs_tq = tensor_quantizer_cls(
        lhs.shape, lhs_config, name="lhs")
    rhs_tq = tensor_quantizer_cls(
        rhs.shape, rhs_config, name="rhs")
    lhs_bwd_tq, rhs_bwd_tq = None, None
    grad_shape = aqt_einsum.get_out_shape(eq, lhs.shape, rhs.shape)
    if lhs_bwd_config:
      lhs_bwd_tq = aqt_tensor.DynamicTensorQuantizer(
          grad_shape, lhs_bwd_config, name="lhs_bwd"
      )
    if rhs_bwd_config:
      rhs_bwd_tq = aqt_tensor.DynamicTensorQuantizer(
          grad_shape, rhs_bwd_config, name="rhs_bwd"
      )

  event_count = tf.constant(0, tf.int64)
  updates = []
  for argument, weights, tq in [
      (lhs, lhs_weights, lhs_tq),
      (rhs, rhs_weights, rhs_tq),
  ]:
    # update statistics only if the quantizer is not dynamic
    if isinstance(tq, aqt_tensor.TensorQuantizer):
      updates.append(tq.update(argument, weights, event_count))
  with tf.control_dependencies(updates):
    return aqt_ops.aqt_einsum(
        eq,
        lhs_tq,
        lhs,
        rhs_tq,
        rhs,
        train,
        quantize_bwd,
        lhs_bwd_tq,
        rhs_bwd_tq,
        event_count=tf.constant(0, tf.int64) if dynamic_fwd_quant else None,
        use_real_int8_einsum=True,
        **einsum_kwargs,
    )


def _generate_missing_shared_axes() -> Sequence[Dict[str, Any]]:
  """Cases where shared axes are missing."""

  keys = ["testcase_name", "eq", "lhs_share", "rhs_share"]
  cases: Sequence[Tuple[str, str, Sequence[int], Sequence[int]]] = [
      ("no_sharing_one_lhs_valid", "i,->i", [], []),
      ("no_sharing_one_rhs_valid", ",i->i", [], []),
      ("no_sharing_both_valid", "i,j->ij", [], []),
      ("contracting_partial_rhs", "j,j->", [], [0]),
      ("contracting_partial_lhs", "j,j->", [0], []),
      ("contracting_valid", "j,j->", [0], [0]),
  ]

  cases_dicts = []
  for vals in cases:
    case = dict(zip(keys, vals))
    case["is_valid"] = typing.cast(str, case["testcase_name"]).endswith("valid")
    cases_dicts.append(case)

  return cases_dicts


def _generate_diag_equation() -> Sequence[Dict[str, Any]]:
  """Cases where shared axes are missing."""

  keys = ["testcase_name", "eq", "lhs_rank", "rhs_rank"]
  cases = [
      ("diag_left", "ii,->i", 2, 0),
      ("diag_right", ",ii->i", 0, 2),
      ("diag_both", "ii,jj->ij", 2, 2),
      ("matmul_and_diag_left", "ii,ik->k", 2, 2),
      ("matmul_and_diag_right", "ij,jj->i", 2, 2),
  ]
  return [dict(zip(keys, vals)) for vals in cases]


def _generate_self_contracting_equation() -> Sequence[Dict[str, Any]]:
  """Cases where shared axes are missing."""

  keys = ["testcase_name", "eq", "lhs_rank", "rhs_rank"]
  cases = [
      ("sum_left", "i,->", 1, 0),
      ("sum_right", ",i->", 0, 1),
      ("sum_both", "i,j->", 1, 1),
      ("batch_matmul_sum_right", "bij,jk->bi", 3, 2),
      ("batch_matmul_sum_left", "bij,jk->bk", 3, 2),
      ("batch_matmul_sum_both", "bij,jk->b", 3, 2),
  ]
  return [dict(zip(keys, vals)) for vals in cases]


def _generate_bad_equation() -> Sequence[Dict[str, Any]]:
  """Cases where shared axes are missing."""

  keys = ["testcase_name", "eq", "lhs_rank", "rhs_rank"]
  cases = [
      ("single_arg", "ii->", 2, 0),
      ("single_arg_no_out", "ii", 2, 0),
      ("double_arg_no_out", "ii,ij", 2, 2),
      ("bad_out", "i,i>i", 2, 2),
      ("bad_out_dash", "i,i-i", 2, 2),
      ("nonstandard_axes", "i!j,ij->i", 3, 2),
      ("space", "i j,ij->i", 3, 2),
      ("newline", "ij,i\nj->i", 2, 3),
      ("ellipses", "...ij,jk->ik", 4, 2),
      ("diag", "ii,->i", 2, 0),
      ("sum", "i,->", 1, 0),
      ("trace", "ii,->", 2, 0),
  ]
  return [dict(zip(keys, vals)) for vals in cases]


def _generate_test_equations(
    append_quantize_bwd: bool = True) -> Sequence[Dict[str, Any]]:
  keys = ["testcase_name", "eq", "quantize_bwd"]
  eq_cases = [
      ("transpose", ",ij->ji"),
      ("matmul", "ij,jk->ik"),
      ("batch_matmul", "bij,bjk->bik"),
      ("dot", "i,i->"),
      ("vec_mat_mult", "u,uv->v"),
      ("channel_vec_mat_mult", "u,nuv->nv"),
      ("batch_channel_vec_mat_mult", "bu,nuv->bnv"),
      ("block_vec_mat_mult", "nu,nuv->nv"),
      ("batch_block_vec_mat_mult", "bnu,nuv->bnv"),
      ("batch_transpose_matmul", "bmu,nm->bnu"),
  ]

  if not append_quantize_bwd:
    return [dict(zip(keys, vals)) for vals in eq_cases]

  cases = []
  for name, eq in eq_cases:
    # append a value indicate whether quantize backward pass
    cases.append((name + "_quantize_bwd", eq, True))
    cases.append((name, eq, False))

  return [dict(zip(keys, vals)) for vals in cases]


def _generate_equations_with_axes() -> Sequence[Dict[str, Any]]:
  keys = ["testcase_name", "eq", "lhs_contracting_axes", "rhs_contracting_axes",
          "lhs_diagnal_axes", "rhs_diagnal_axes", "lhs_self_contracting_labels",
          "rhs_self_contracting_labels"]
  cases = [
      ("diag", "ii,->i", [], [], {"i": [0, 1]}, {}, [], []),
      ("sum", "i,->", [0], [], {}, {}, ["i"], []),
      ("trace", "ii,->", [0, 1], [], {"i": [0, 1]}, {}, ["i"], []),
      ("transpose", ",ij->ji", [], [], {}, {}, [], []),
      ("matmul", "ij,jk->ik", [1], [0], {}, {}, [], []),
      ("batch_matmul", "bij,bjk->bik", [2], [1], {}, {}, [], []),
      ("dot", "i,i->", [0], [0], {}, {}, [], []),
      ("vec_mat_mult", "u,uv->v", [0], [0], {}, {}, [], []),
      ("channel_vec_mat_mult", "u,nuv->nv", [0], [1], {}, {}, [], []),
      ("batch_channel_vec_mat_mult", "bu,nuv->bnv", [1], [1], {}, {}, [], []),
      ("block_vec_mat_mult", "nu,nuv->nv", [1], [1], {}, {}, [], []),
      ("batch_block_vec_mat_mult", "bnu,nuv->bnv", [2], [1], {}, {}, [], []),
      ("batch_transpose_matmul", "bmu,nm->bnu", [1], [1], {}, {}, [], []),
  ]
  return [dict(zip(keys, vals)) for vals in cases]


class EinsumTest(tf.test.TestCase, parameterized.TestCase):

  def setUp(self):
    """Seed random for deterministic but nontrivial inputs."""
    super().setUp()
    self.rng = np.random.default_rng(1234)

  def randn(self, *size):
    return self.rng.standard_normal(size=size, dtype=np.float32)

  @parameterized.named_parameters(_generate_missing_shared_axes())
  def test_missing_shared_axes(
      self,  #
      eq: str,
      lhs_share: Sequence[int],
      rhs_share: Sequence[int],
      is_valid: bool):

    def make_tensor(einsum_str):
      return tf.constant(np.ones([1] * len(einsum_str)), tf.float32)

    def make_op():
      lhs, rhs, _ = aqt_einsum._parse_equation(eq)
      lhs, rhs = make_tensor(lhs), make_tensor(rhs)
      return _einsum_op(eq, lhs, rhs, _empty_config(lhs_share),
                        _empty_config(rhs_share))

    if is_valid:
      make_op()
    else:
      with self.assertRaisesRegex(aqt_config.ConfigError,
                                  "axis .* of .* must be shared due to .*"):
        make_op()

  @parameterized.named_parameters(_generate_diag_equation())
  def test_diag_equation(self, eq: str, lhs_rank: int, rhs_rank: int):
    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "expected to have no diagnalization"):

      def make_tensor(rank):
        return tf.constant(np.ones([1] * rank), tf.float32)

      lhs = make_tensor(lhs_rank)
      rhs = make_tensor(rhs_rank)
      lhs_config = _empty_config(list(range(lhs_rank)))
      rhs_config = _empty_config(list(range(rhs_rank)))
      _einsum_op(eq, lhs, rhs, lhs_config, rhs_config)

  @parameterized.named_parameters(_generate_self_contracting_equation())
  def test_self_contracting_equation(self, eq: str, lhs_rank: int,
                                     rhs_rank: int):
    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "expected to have no self-contraction"):

      def make_tensor(rank):
        return tf.constant(np.ones([1] * rank), tf.float32)

      lhs = make_tensor(lhs_rank)
      rhs = make_tensor(rhs_rank)
      lhs_config = _empty_config(list(range(lhs_rank)))
      rhs_config = _empty_config(list(range(rhs_rank)))
      _einsum_op(eq, lhs, rhs, lhs_config, rhs_config)

  @parameterized.named_parameters(_generate_bad_equation())
  def test_bad_equation(self, eq: str, lhs_rank: int, rhs_rank: int):
    with self.assertRaisesRegex(aqt_config.ConfigError, "einsum equation"):

      def make_tensor(rank):
        return tf.constant(np.ones([1] * rank), tf.float32)

      lhs = make_tensor(lhs_rank)
      rhs = make_tensor(rhs_rank)
      lhs_config = _empty_config(list(range(lhs_rank)))
      rhs_config = _empty_config(list(range(rhs_rank)))
      _einsum_op(eq, lhs, rhs, lhs_config, rhs_config)

  def basic_quant_example(self):
    eq = "bji,jk->bik"

    # Representable values: -6, -4, -2, 0, 2, 4, 6 (preserving zeros).
    # Data batch axis: 0
    # Contracting axis: 1
    lhs_bound = 7
    lhs_config = _schedule_config(3, lhs_bound, [0, 1])
    lhs = tf.constant(
        np.array(
            [
                [
                    [-8, -5.99],  #
                    [4.01, 0.01],
                    [4.01, -4.01]
                ],
                [
                    [-0.01, 2.01],  #
                    [4.01, 6.01],
                    [3.99, -3.99]
                ]
            ],
            dtype=np.float32))
    qlhs_value = tf.constant(
        np.array(
            [
                [
                    [-6, -6],  #
                    [4, 0],
                    [4, -4]
                ],
                [
                    [0, 2],  #
                    [4, 6],
                    [4, -4]
                ]
            ],
            dtype=np.float32))

    # manual clip and quantization to be differentiable.
    qlhs = tf.clip_by_value(lhs, -lhs_bound, lhs_bound)
    qlhs = tf.grad_pass_through(lambda x: qlhs_value)(qlhs)

    # Representable values: -1, 0, 1
    # Contracting axis: 0
    rhs_bound = 1.5
    rhs_config = _schedule_config(2, 1.5, [0])
    rhs = tf.constant(
        np.array(
            [
                [-3, 0.99],  #
                [-0.99, 0],
                [-0.01, 2]
            ],
            dtype=np.float32))
    qrhs_value = tf.constant(
        np.array(
            [
                [-1, 1],  #
                [-1, 0],
                [0, 1]
            ],
            dtype=np.float32))

    # similar manual clip and quantization but with a different clip bound
    qrhs = tf.clip_by_value(rhs, -rhs_bound, rhs_bound)
    qrhs = tf.grad_pass_through(lambda x: qrhs_value)(qrhs)

    return eq, lhs_config, lhs, qlhs, rhs_config, rhs, qrhs

  def basic_emulation_example(self):
    eq = "ij,jk->ik"
    # Data batch axis: 0
    # Contracting axis: 1
    lhs_config = _schedule_config_emulation([0, 1])
    lhs = tf.constant(
        np.array(
            [
                [-8.5, 4.3, 4.1],  #
                [-0.05, 0.01, -4.7],
            ],
            dtype=np.float32))
    qlhs = tf.constant(
        np.array(
            [
                [-8.0, 4.0, 4.0],  #
                [-0.046875, 0.00976562, -5.0]
            ],
            dtype=np.float32))

    # Contracting axis: 0
    rhs_config = _schedule_config_emulation([0])
    rhs = tf.constant(
        np.array(
            [
                [-0.2, 0.02],  #
                [-1.1, 0],
                [-0.04, 2.3]
            ],
            dtype=np.float32))
    qrhs = tf.constant(
        np.array(
            [
                [-0.1875, 0.01953125],  #
                [-1.0, 0.0],
                [-0.0390625, 2.5]
            ],
            dtype=np.float32))

    return eq, lhs_config, lhs, qlhs, rhs_config, rhs, qrhs

  def test_basic_einsum(self):
    with self.subTest("quant_example"):
      eq, lhs_config, lhs, qlhs, rhs_config, rhs, qrhs = (
          self.basic_quant_example())

    with self.subTest("emulation_example"):
      eq, lhs_config, lhs, qlhs, rhs_config, rhs, qrhs = (
          self.basic_emulation_example())

    actual = _einsum_op(eq, lhs, rhs, lhs_config, rhs_config)
    expected = tf.einsum(eq, qlhs, qrhs)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      self.assertAllEqual(expected, actual)

  @parameterized.named_parameters(
      {"testcase_name": "no_bwd_quant", "quantize_bwd": False},
      {"testcase_name": "bwd_quant", "quantize_bwd": True},
  )
  def test_no_quantization(self, quantize_bwd: bool):
    lhs = tf.constant(self.randn(3, 4))
    rhs = tf.constant(self.randn(4, 2))
    eq = "ij,jk->ik"
    lhs_config = _empty_config([1])
    rhs_config = _empty_config([0])

    lhs_float_config = _schedule_config(8, 1.0, [1])
    rhs_float_config = _schedule_config(8, 1.0, [0])

    lhs_float_config.tensor_configs[0].quant_config = aqt_config.FloatConfig()
    rhs_float_config.tensor_configs[0].quant_config = aqt_config.FloatConfig()

    def _get_grad_config(eq: str,
                         swap_ans: bool
                         ) -> Optional[aqt_config.AqtScheduleConfig]:
      if not quantize_bwd:
        return None
      bwd_eq = aqt_einsum.get_einsum_transpose(eq, swap_ans=swap_ans)
      grad_caxes, _ = aqt_einsum.get_contracting_axes(bwd_eq)
      # 16 bits to preserve gradients
      grad_config = _schedule_config(8, 1.0, grad_caxes,
                                     freeze_scale_at_begin=False)
      grad_config.use_quantized_variable = False
      grad_config.tensor_configs[0].quant_config = aqt_config.FloatConfig()
      return grad_config

    lhs_bwd_config = _get_grad_config(eq, False)
    rhs_bwd_config = _get_grad_config(eq, True)

    actual_fwd = _einsum_op(
        eq, lhs, rhs, lhs_config, rhs_config, varscope_name="no_quant_einsum",
        quantize_bwd=quantize_bwd, lhs_bwd_config=lhs_bwd_config,
        rhs_bwd_config=rhs_bwd_config)
    float_config_actual_fwd = _einsum_op(
        eq, lhs, rhs, lhs_float_config, rhs_float_config,
        varscope_name="float_config_einsum",
        quantize_bwd=quantize_bwd, lhs_bwd_config=lhs_bwd_config,
        rhs_bwd_config=rhs_bwd_config,
        )
    expected_fwd = tf.einsum(eq, lhs, rhs)

    actual_lgrad, actual_rgrad = tf.gradients([actual_fwd], [lhs, rhs])
    float_config_actual_lgrad, float_config_actual_rgrad = tf.gradients(
        [float_config_actual_fwd], [lhs, rhs])
    expected_lgrad, expected_rgrad = tf.gradients([expected_fwd], [lhs, rhs])

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      with self.subTest("fwd_int"):
        self.assertAllEqual(expected_fwd, actual_fwd)
      with self.subTest("fwd_float"):
        self.assertAllEqual(expected_fwd, float_config_actual_fwd)
      with self.subTest("bwd_int"):
        self.assertAllEqual(actual_lgrad, expected_lgrad)
        self.assertAllEqual(actual_rgrad, expected_rgrad)
      with self.subTest("bwd_float"):
        self.assertAllEqual(float_config_actual_lgrad, expected_lgrad)
        self.assertAllEqual(float_config_actual_rgrad, expected_rgrad)

  @parameterized.parameters([dict(lhs_float=True), dict(lhs_float=False)])
  def test_float_config_basic_einsum(self, lhs_float):
    eq, lhs_config, lhs, qlhs, rhs_config, rhs, qrhs = self.basic_quant_example(
    )
    if lhs_float:
      lhs_config.tensor_configs[0].quant_config = aqt_config.FloatConfig()
    else:
      rhs_config.tensor_configs[0].quant_config = aqt_config.FloatConfig()

    actual_fwd = _einsum_op(eq, lhs, rhs, lhs_config, rhs_config)
    if lhs_float:
      qlhs = lhs  # lhs is not quantized
    else:
      qrhs = rhs  # rhs is not quantized
    expected_fwd = tf.einsum(eq, qlhs, qrhs)

    actual_lgrad, actual_rgrad = tf.gradients([actual_fwd], [lhs, rhs])
    expected_lgrad, expected_rgrad = tf.gradients([expected_fwd], [lhs, rhs])

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      with self.subTest("fwd"):
        self.assertAllEqual(expected_fwd, actual_fwd)
      with self.subTest("bwd"):
        self.assertAllEqual(actual_lgrad, expected_lgrad)
        self.assertAllEqual(actual_rgrad, expected_rgrad)

  def test_passes_arguments_to_inner_einsum(self):
    module = "tensorflow.compat.v1"
    with mock.patch(f"{module}.einsum", side_effect=tf.einsum) as tfeinsum:
      lhs = tf.constant(self.randn(3, 4))
      rhs = tf.constant(self.randn(4, 2))
      eq = "ij,jk->ik"
      lhs_config = _schedule_config(8, 1.0, [1])
      rhs_config = _schedule_config(8, 1.0, [0])

      kwargs = {"optimize": "optimal"}

      _einsum_op(eq, lhs, rhs, lhs_config, rhs_config, **kwargs)
      for (_, actual_kwargs) in tfeinsum.call_args_list:
        subset = {k: v for k, v in actual_kwargs.items() if k in kwargs}
        self.assertEqual(subset, kwargs)

  @parameterized.named_parameters(
      aqt_test_shared_base.generate_unaligned_schedule_intervals())
  def test_unaligned_schedule_intervals(self, lhs_intervals, rhs_intervals):

    def config_from_schedule(intervals,
                             share_stats_axes):
      config = _empty_config(share_stats_axes)
      for start, stop in intervals:
        config.tensor_configs += _schedule_config(
            8, 1.0, share_stats_axes).tensor_configs
        config.tensor_configs[-1].begin_at_event = start
        config.tensor_configs[-1].end_at_event = stop
      return config

    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "intervals do not match|config len"):
      lhs = tf.constant(self.randn(3, 4))
      rhs = tf.constant(self.randn(4, 2))
      eq = "ij,jk->ik"
      # share_stats_axes includes the batch axies i and the contracting axis j.
      lhs_config = config_from_schedule(lhs_intervals, [0, 1])
      rhs_config = config_from_schedule(rhs_intervals, [0])
      _einsum_op(eq, lhs, rhs, lhs_config, rhs_config)

  def exact_int8_einsum_example(
      self,
      eq: str,
      quantize_lhs: bool = False,
      quantize_rhs: bool = False,
      scale: float = 1.0,
      quantize_bwd: bool = False,
      dynamic_bwd_quant: bool = False,
      dynamic_fwd_quant: bool = False,
  ):
    """Returns a pair of tensors and config to einsum exactly."""
    lhs, rhs, _ = aqt_einsum._parse_equation(eq)

    # A subset of the range of numbers which can be preserved exactly.
    bits = 8
    symmetric_uniform_range = 2**(bits - 1) - 1
    lo, hi = -symmetric_uniform_range, symmetric_uniform_range

    axis_labels = sorted(set(lhs + rhs))
    label_dims = {k: self.rng.integers(2, 10) for k in axis_labels}
    lhs_shape = [label_dims[k] for k in lhs]
    rhs_shape = [label_dims[k] for k in rhs]

    def make_tensor(shape):
      np_tensor = self.rng.integers(lo, hi, size=shape, dtype=np.int64)
      return tf.constant(np_tensor, dtype=tf.float32)

    lhs = make_tensor(lhs_shape) * scale
    rhs = make_tensor(rhs_shape) * scale

    def _exact_schedule_config(bits, eq, scale, freeze_scale):
      iqc = aqt_config.IntQuantConfig(bits=bits, preserve_zero=True)
      clip_bound = aqt_common.get_clip_bound(iqc)
      assert symmetric_uniform_range <= clip_bound

      lhs_caxes, rhs_caxes = aqt_einsum.get_contracting_axes(eq)

      # to exactly represent quantized lhs and rhs
      const_bound_coeff = scale * clip_bound
      lhs_config = _schedule_config(bits, const_bound_coeff, lhs_caxes,
                                    freeze_scale_at_begin=freeze_scale)
      rhs_config = _schedule_config(bits, const_bound_coeff, rhs_caxes,
                                    freeze_scale_at_begin=freeze_scale)
      return lhs_config, rhs_config

    lhs_config, rhs_config = _exact_schedule_config(8, eq, scale,
                                                    not dynamic_fwd_quant)

    lhs_config.use_quantized_variable = quantize_lhs and not dynamic_fwd_quant
    rhs_config.use_quantized_variable = quantize_rhs and not dynamic_fwd_quant

    def _get_grad_config(eq: str,
                         swap_ans: bool
                         ) -> Optional[aqt_config.AqtScheduleConfig]:
      if not quantize_bwd:
        return None
      bwd_eq = aqt_einsum.get_einsum_transpose(eq, swap_ans=swap_ans)
      # 16 bits to preserve gradients
      grad_config, _ = _exact_schedule_config(16, bwd_eq, 1.0, False)
      grad_config.use_quantized_variable = False
      for tc in grad_config.tensor_configs:
        tc.freeze_scale_at_begin = False
      return grad_config

    lhs_bwd_config = _get_grad_config(eq, False)
    rhs_bwd_config = _get_grad_config(eq, True)

    # Change calibration configs if dynamic quant in the backward pass
    if lhs_bwd_config and dynamic_bwd_quant:
      for tc in lhs_bwd_config.tensor_configs:
        tc.calibration_config = aqt_config.CalibrationConfig(
            const_bound_coeff=1.0,
            max_dev_coeff=1.0,
        )
    if rhs_bwd_config and dynamic_bwd_quant:
      for tc in rhs_bwd_config.tensor_configs:
        tc.calibration_config = aqt_config.CalibrationConfig(
            const_bound_coeff=1.0,
            max_dev_coeff=1.0,
        )

    return lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config

  @parameterized.named_parameters(_generate_test_equations())
  def test_vars_dont_kill_grads(self, eq, quantize_bwd):
    with self.subTest("scale_one"):
      lhs_config, lhs, rhs_config, rhs, lbwd_config, rbwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=1.0, quantize_bwd=quantize_bwd,
          )
      )
      lhs_config_novar, _, rhs_config_novar, _, lbwd_config, rbwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=1.0, quantize_bwd=quantize_bwd
          )
      )
    with self.subTest("scale_two"):
      lhs_config, lhs, rhs_config, rhs, lbwd_config, rbwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=2.0, quantize_bwd=quantize_bwd
          )
      )
      lhs_config_novar, _, rhs_config_novar, _, lbwd_config, rbwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=2.0, quantize_bwd=quantize_bwd
          )
      )

    expected_op = _einsum_op(
        eq,
        lhs,
        rhs,
        lhs_config_novar,
        rhs_config_novar,
        varscope_name="novar",
        quantize_bwd=quantize_bwd,
        lhs_bwd_config=lbwd_config,
        rhs_bwd_config=rbwd_config,
    )
    actual_op = _einsum_op(
        eq,
        lhs,
        rhs,
        lhs_config,
        rhs_config,
        varscope_name="var",
        quantize_bwd=quantize_bwd,
        lhs_bwd_config=lbwd_config,
        rhs_bwd_config=rbwd_config,
    )

    saved_grads = tf.gradients([expected_op], [lhs, rhs])
    unsaved_grads = tf.gradients([actual_op], [lhs, rhs])

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()

      zipped_grads = zip(saved_grads, unsaved_grads)
      for actual_grad, expected_grad in zipped_grads:
        self.assertAllEqual(actual_grad, expected_grad)

  @parameterized.named_parameters(_generate_test_equations())
  def test_vars_over_inputs_at_inference(self, eq, quantize_bwd):
    with self.subTest("scale_one"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=1.0, quantize_bwd=quantize_bwd
          )
      )
    with self.subTest("scale_two"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=2.0, quantize_bwd=quantize_bwd
          )
      )

    lhs_tq = aqt_tensor.TensorQuantizer(lhs.shape, lhs_config, name="lhs")
    rhs_tq = aqt_tensor.TensorQuantizer(rhs.shape, rhs_config, name="rhs")
    if quantize_bwd:
      grad_shape = aqt_einsum.get_out_shape(eq, lhs.shape, rhs.shape)
      lhs_bwd_tq = aqt_tensor.DynamicTensorQuantizer(
          grad_shape, lhs_bwd_config, name="lhs_bwd"
      )
      rhs_bwd_tq = aqt_tensor.DynamicTensorQuantizer(
          grad_shape, rhs_bwd_config, name="rhs_bwd"
      )
    else:
      lhs_bwd_tq = rhs_bwd_tq = None

    # Update at least once to initialize scale, then grab the expected
    # value while in training mode.
    event_count = tf.constant(0, tf.int64)
    updates = [
        lhs_tq.update(lhs, weight=None, event_count=event_count),
        rhs_tq.update(rhs, weight=None, event_count=event_count)
    ]
    with tf.control_dependencies(updates):
      expected = aqt_ops.aqt_einsum(
          eq,
          lhs_tq,
          lhs,
          rhs_tq,
          rhs,
          train=True,
          quantize_bwd=quantize_bwd,
          lhs_grad_quantizer=lhs_bwd_tq,
          rhs_grad_quantizer=rhs_bwd_tq,
      )

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      expected = expected.eval()

      actual = aqt_ops.aqt_einsum(
          eq,
          lhs_tq,
          tf.zeros_like(lhs),
          rhs_tq,
          tf.zeros_like(rhs),
          train=False,
          quantize_bwd=quantize_bwd,
          lhs_grad_quantizer=lhs_bwd_tq,
          rhs_grad_quantizer=rhs_bwd_tq,
      )

      self.assertAllEqual(actual, expected)

  @parameterized.named_parameters(_generate_test_equations())
  def test_float_config_not_save_quantized_var(self, eq, quantize_bwd):
    with self.subTest("scale_one"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=1.0, quantize_bwd=quantize_bwd
          )
      )
    with self.subTest("scale_two"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=2.0, quantize_bwd=quantize_bwd
          )
      )

    lhs_config.tensor_configs[0].quant_config = aqt_config.FloatConfig()
    lhs_tq = aqt_tensor.TensorQuantizer(lhs.shape, lhs_config, name="lhs")
    rhs_tq = aqt_tensor.TensorQuantizer(rhs.shape, rhs_config, name="rhs")
    if quantize_bwd:
      grad_shape = aqt_einsum.get_out_shape(eq, lhs.shape, rhs.shape)
      lhs_bwd_tq = aqt_tensor.DynamicTensorQuantizer(
          grad_shape, lhs_bwd_config, name="lhs_bwd"
      )
      rhs_bwd_tq = aqt_tensor.DynamicTensorQuantizer(
          grad_shape, rhs_bwd_config, name="rhs_bwd"
      )
    else:
      lhs_bwd_tq = rhs_bwd_tq = None

    event_count = tf.constant(0, tf.int64)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      lhs_tq.update(lhs, weight=None, event_count=event_count).run()
      rhs_tq.update(rhs, weight=None, event_count=event_count).run()

      actual = aqt_ops.aqt_einsum(
          eq,
          lhs_tq,
          lhs,
          rhs_tq,
          rhs,
          train=False,
          quantize_bwd=quantize_bwd,
          lhs_grad_quantizer=lhs_bwd_tq,
          rhs_grad_quantizer=rhs_bwd_tq,
      )
      # Although the input tensors are non-zeros, the result of einsum with
      # inference mode should be zeros because lhs uses zero-initialized
      # quantized var while rhs can restore its updated quantized variable.
      expected = tf.zeros_like(actual)

      self.assertAllEqual(actual, expected)

  @parameterized.named_parameters(_generate_test_equations())
  def test_exact(self, eq, quantize_bwd):
    with self.subTest("scale_one"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=1.0, quantize_bwd=quantize_bwd
          )
      )
    with self.subTest("scale_two"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=2.0, quantize_bwd=quantize_bwd
          )
      )

    actual = _einsum_op(
        eq,
        lhs,
        rhs,
        lhs_config,
        rhs_config,
        quantize_bwd=quantize_bwd,
        lhs_bwd_config=lhs_bwd_config,
        rhs_bwd_config=rhs_bwd_config,
    )
    expected = tf.einsum(eq, lhs, rhs)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      self.assertAllEqual(actual, expected)

  @parameterized.named_parameters(_generate_test_equations())
  def test_exact_grads(self, eq, quantize_bwd):
    with self.subTest("scale_one"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=1.0, quantize_bwd=quantize_bwd,
          )
      )
    with self.subTest("scale_two"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, True, True, scale=2.0, quantize_bwd=quantize_bwd,
          )
      )

    actual_fwd = _einsum_op(
        eq,
        lhs,
        rhs,
        lhs_config,
        rhs_config,
        quantize_bwd=quantize_bwd,
        lhs_bwd_config=lhs_bwd_config,
        rhs_bwd_config=rhs_bwd_config,
    )
    expected_fwd = tf.einsum(eq, lhs, rhs)

    expected = tf.gradients([expected_fwd], [lhs, rhs])
    actual = tf.gradients([actual_fwd], [lhs, rhs])

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      for actual_grad, expected_grad in zip(actual, expected):
        # Uniform noises in [-5, 5] while {-5, 5} has zero measure.
        # Expect the gradient is still exact since (-5, 5) does not change
        # rouding.
        self.assertAllEqual(actual_grad, expected_grad)

  @parameterized.named_parameters(_generate_test_equations())
  def test_exact_grads_dynamic(self, eq, quantize_bwd):
    lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
        self.exact_int8_einsum_example(
            eq, True, True, scale=2.0, quantize_bwd=quantize_bwd,
            dynamic_fwd_quant=True
        )
    )

    actual_fwd = _einsum_op(
        eq,
        lhs,
        rhs,
        lhs_config,
        rhs_config,
        quantize_bwd=quantize_bwd,
        lhs_bwd_config=lhs_bwd_config,
        rhs_bwd_config=rhs_bwd_config,
        dynamic_fwd_quant=True,
    )
    expected_fwd = tf.einsum(eq, lhs, rhs)

    expected = tf.gradients([expected_fwd], [lhs, rhs])
    actual = tf.gradients([actual_fwd], [lhs, rhs])

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      with self.subTest("fwd"):
        self.assertAllEqual(actual_fwd, expected_fwd)
      with self.subTest("bwd"):
        for actual_grad, expected_grad in zip(actual, expected):
          # Uniform noises in [-5, 5] while {-5, 5} has zero measure.
          # Expect the gradient is still exact since (-5, 5) does not change
          # rouding.
          self.assertAllEqual(actual_grad, expected_grad)

  @parameterized.named_parameters(_generate_test_equations())
  def test_inexact(self, eq, quantize_bwd):
    with self.subTest("larger_x_bound"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, scale=2.0, quantize_bwd=quantize_bwd
              )
      )
      lhs, rhs = lhs / 2.0, rhs / 2.0
    with self.subTest("smaller_x_bound"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, scale=1.0, quantize_bwd=quantize_bwd
              )
      )
      lhs, rhs = lhs * 2.0, rhs * 2.0

    actual = _einsum_op(
        eq,
        lhs,
        rhs,
        lhs_config,
        rhs_config,
        quantize_bwd=quantize_bwd,
        lhs_bwd_config=lhs_bwd_config,
        rhs_bwd_config=rhs_bwd_config,
    )
    expected = tf.einsum(eq, lhs, rhs)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      self.assertNotAllEqual(actual, expected)

  @parameterized.named_parameters(_generate_test_equations())
  def test_inexact_grads(self, eq, quantize_bwd):
    with self.subTest("larger_x_bound"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, scale=2.0, quantize_bwd=quantize_bwd
          )
      )
      lhs, rhs = lhs / 2.0, rhs / 2.0
    with self.subTest("smaller_x_bound"):
      lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
          self.exact_int8_einsum_example(
              eq, scale=1.0, quantize_bwd=quantize_bwd
          )
      )
      lhs, rhs = lhs * 2.0, rhs * 2.0

    actual_fwd = _einsum_op(
        eq,
        lhs,
        rhs,
        lhs_config,
        rhs_config,
        quantize_bwd=quantize_bwd,
        lhs_bwd_config=lhs_bwd_config,
        rhs_bwd_config=rhs_bwd_config,
    )
    expected_fwd = tf.einsum(eq, lhs, rhs)

    expected = tf.gradients([expected_fwd], [lhs, rhs])
    actual = tf.gradients([actual_fwd], [lhs, rhs])

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      for actual_grad, expected_grad in zip(actual, expected):
        self.assertNotAllEqual(actual_grad, expected_grad)

  @parameterized.named_parameters(_generate_test_equations(
      # Testing bwd quant only, no need to include quantize_bwd in test cases.
      append_quantize_bwd=False,
      ))
  def test_consistent_bwd_improves_grads(self, eq):
    lhs_config, lhs, rhs_config, rhs, lhs_bwd_config, rhs_bwd_config = (
        self.exact_int8_einsum_example(
            eq, quantize_bwd=True, dynamic_bwd_quant=True,
        )
    )
    def get_perturbed_gradients(step_i):
      actual_fwd = _einsum_op(
          eq,
          lhs,
          rhs,
          lhs_config,
          rhs_config,
          quantize_bwd=True,
          lhs_bwd_config=lhs_bwd_config,
          rhs_bwd_config=rhs_bwd_config,
          varscope_name=f"einsum_seed_{step_i}",
      )
      return tf.gradients([actual_fwd], [lhs, rhs])

    exact_fwd = tf.einsum(eq, lhs, rhs)
    exact = tf.gradients([exact_fwd], [lhs, rhs])

    num_samples = 8
    qgrad_samples = [get_perturbed_gradients(i) for i in range(num_samples)]
    estimate1 = qgrad_samples[0]
    estimate8 = [tf.reduce_mean(g, axis=0) for g in zip(*qgrad_samples[:8])]

    def get_error(estimate):
      return [tf.linalg.norm(i - j) for i, j in zip(estimate, exact)]

    sample_errors = get_error(estimate1)
    ensemble_errors = get_error(estimate8)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()

      for estimate1_g, exact_g, sample_error, ensemble_err in zip(
          estimate1, exact, sample_errors, ensemble_errors
      ):
        # Check dynamic backward quant should be close
        self.assertAllClose(estimate1_g, exact_g, rtol=1e-2)
        # the unbiased estimate should eventually converge or make improvement
        self.assertAllLess(ensemble_err, sample_error)

  @parameterized.named_parameters(_generate_equations_with_axes())
  def test_equations_with_axes(self, eq,
                               lhs_contracting_axes, rhs_contracting_axes,
                               lhs_diagnal_axes, rhs_diagnal_axes,
                               lhs_self_contracting_labels,
                               rhs_self_contracting_labels):
    l_axes, r_axes = aqt_einsum.get_contracting_axes(eq)
    for actual, expected in [(l_axes, lhs_contracting_axes),
                             (r_axes, rhs_contracting_axes)]:
      self.assertAllEqual(actual, expected)
    lhs, rhs, out = aqt_einsum._parse_equation(eq)
    for actual, expected in [(aqt_einsum._get_diagnal_axes(lhs),
                              lhs_diagnal_axes),
                             (aqt_einsum._get_diagnal_axes(rhs),
                              rhs_diagnal_axes)]:
      self.assertAllEqual(actual, expected)
    lc, rc = aqt_einsum._get_self_contracting_labels(lhs, rhs, out)
    for actual, expected in [(lc, lhs_self_contracting_labels),
                             (rc, rhs_self_contracting_labels)]:
      self.assertAllEqual(actual, expected)

if __name__ == "__main__":
  absltest.main()
